# -*- coding: utf-8 -*-
#
# This file is part of the python-chess library.
# Copyright (C) 2012-2016 Niklas Fiekas <niklas.fiekas@backscattering.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import chess
import itertools
import re
import logging

try:
    import backport_collections as collections
except ImportError:
    import collections


LOGGER = logging.getLogger(__name__)


NAG_NULL = 0

NAG_GOOD_MOVE = 1
"""A good move. Can also be indicated by ``!`` in PGN notation."""

NAG_MISTAKE = 2
"""A mistake. Can also be indicated by ``?`` in PGN notation."""

NAG_BRILLIANT_MOVE = 3
"""A brilliant move. Can also be indicated by ``!!`` in PGN notation."""

NAG_BLUNDER = 4
"""A blunder. Can also be indicated by ``??`` in PGN notation."""

NAG_SPECULATIVE_MOVE = 5
"""A speculative move. Can also be indicated by ``!?`` in PGN notation."""

NAG_DUBIOUS_MOVE = 6
"""A dubious move. Can also be indicated by ``?!`` in PGN notation."""

NAG_FORCED_MOVE = 7
NAG_SINGULAR_MOVE = 8
NAG_WORST_MOVE = 9
NAG_DRAWISH_POSITION = 10
NAG_QUIET_POSITION = 11
NAG_ACTIVE_POSITION = 12
NAG_UNCLEAR_POSITION = 13
NAG_WHITE_SLIGHT_ADVANTAGE = 14
NAG_BLACK_SLIGHT_ADVANTAGE = 15

# TODO: Add more constants for example from
# https://en.wikipedia.org/wiki/Numeric_Annotation_Glyphs

NAG_WHITE_MODERATE_COUNTERPLAY = 132
NAG_BLACK_MODERATE_COUNTERPLAY = 133
NAG_WHITE_DECISIVE_COUNTERPLAY = 134
NAG_BLACK_DECISIVE_COUNTERPLAY = 135
NAG_WHITE_MODERATE_TIME_PRESSURE = 136
NAG_BLACK_MODERATE_TIME_PRESSURE = 137
NAG_WHITE_SEVERE_TIME_PRESSURE = 138
NAG_BLACK_SEVERE_TIME_PRESSURE = 139


TAG_REGEX = re.compile(r"^\[([A-Za-z0-9_]+)\s+\"(.*)\"\]\s*$")

MOVETEXT_REGEX = re.compile(r"""
    (%.*?[\n\r])
    |(\{.*)
    |(\$[0-9]+)
    |(\()
    |(\))
    |(\*|1-0|0-1|1/2-1/2)
    |(
        [NBKRQ]?[a-h]?[1-8]?[\-x]?[a-h][1-8](?:=?[nbrqkNBRQK])?
        |[PNBRQK]?@[a-h][1-8]
        |--
        |O-O(?:-O)?
        |0-0(?:-0)?
    )
    |([\?!]{1,2})
    """, re.DOTALL | re.VERBOSE)


class GameNode(object):

    def __init__(self):
        self.parent = None
        self.move = None
        self.nags = set()
        self.starting_comment = ""
        self.comment = ""
        self.variations = []

        self.board_cached = None

    def board(self, _cache=True):
        """
        Gets a board with the position of the node.

        It's a copy, so modifying the board will not alter the game.
        """
        if self.board_cached:
            return self.board_cached.copy()

        board = self.parent.board(_cache=False)
        board.push(self.move)

        if _cache:
            self.board_cached = board
            return board.copy()
        else:
            return board

    def san(self):
        """
        Gets the standard algebraic notation of the move leading to this node.

        Do not call this on the root node.
        """
        return self.parent.board().san(self.move)

    def root(self):
        """Gets the root node, i.e. the game."""
        node = self

        while node.parent:
            node = node.parent

        return node

    def end(self):
        """Follows the main variation to the end and returns the last node."""
        node = self

        while node.variations:
            node = node.variations[0]

        return node

    def is_end(self):
        """Checks if this node is the last node in the current variation."""
        return not self.variations

    def starts_variation(self):
        """
        Checks if this node starts a variation (and can thus have a starting
        comment). The root node does not start a variation and can have no
        starting comment.
        """
        if not self.parent or not self.parent.variations:
            return False

        return self.parent.variations[0] != self

    def is_main_line(self):
        """Checks if the node is in the main line of the game."""
        node = self

        while node.parent:
            parent = node.parent

            if not parent.variations or parent.variations[0] != node:
                return False

            node = parent

        return True

    def is_main_variation(self):
        """
        Checks if this node is the first variation from the point of view of its
        parent. The root node also is in the main variation.
        """
        if not self.parent:
            return True

        return not self.parent.variations or self.parent.variations[0] == self

    def variation(self, move):
        """
        Gets a child node by move or index.
        """
        for index, variation in enumerate(self.variations):
            if move == variation.move or index == move or move == variation:
                return variation

        raise KeyError("variation not found")

    def has_variation(self, move):
        """Checks if the given move appears as a variation."""
        return move in (variation.move for variation in self.variations)

    def promote_to_main(self, move):
        """Promotes the given move to the main variation."""
        variation = self.variation(move)
        self.variations.remove(variation)
        self.variations.insert(0, variation)

    def promote(self, move):
        """Moves the given variation one up in the list of variations."""
        variation = self.variation(move)
        i = self.variations.index(variation)
        if i > 0:
            self.variations[i - 1], self.variations[i] = self.variations[i], self.variations[i - 1]

    def demote(self, move):
        """Moves the given variation one down in the list of variations."""
        variation = self.variation(move)
        i = self.variations.index(variation)
        if i < len(self.variations) - 1:
            self.variations[i + 1], self.variations[i] = self.variations[i], self.variations[i + 1]

    def remove_variation(self, move):
        """Removes a variation by move."""
        self.variations.remove(self.variation(move))

    def add_variation(self, move, comment="", starting_comment="", nags=()):
        """Creates a child node with the given attributes."""
        node = GameNode()
        node.move = move
        node.nags = set(nags)
        node.parent = self
        node.comment = comment
        node.starting_comment = starting_comment
        self.variations.append(node)
        return node

    def add_main_variation(self, move, comment=""):
        """
        Creates a child node with the given attributes and promotes it to the
        main variation.
        """
        node = self.add_variation(move, comment=comment)
        self.variations.remove(node)
        self.variations.insert(0, node)
        return node

    def main_line(self):
        """Yields the moves of the main line starting in this node."""
        node = self
        while node.variations:
            node = node.variations[0]
            yield node.move

    def add_line(self, moves, comment="", starting_comment="", nags=()):
        """
        Creates a sequence of child nodes for the given list of moves.
        Adds *comment* and *nags* to the last node of the line and returns it.
        """
        node = self

        # Add line.
        for move in moves:
            node = node.add_variation(move, starting_comment=starting_comment)
            starting_comment = ""

        # Merge comment and NAGs.
        if node.comment:
            node.comment += " " + comment
        else:
            node.comment = comment

        node.nags.update(nags)

        return node

    def accept(self, visitor, _board=None):
        """
        Traverse game nodes in PGN order using the given *visitor*. Returns
        the visitor result.
        """
        board = self.board() if _board is None else _board

        # The mainline move goes first.
        if self.variations:
            main_variation = self.variations[0]
            visitor.visit_move(board, main_variation.move)

            # Visit NAGs.
            for nag in sorted(main_variation.nags):
                visitor.visit_nag(nag)

            # Visit the comment.
            if main_variation.comment:
                visitor.visit_comment(main_variation.comment)

        # Then visit sidelines.
        for variation in itertools.islice(self.variations, 1, None):
            # Start variation.
            visitor.begin_variation()

            # Append starting comment.
            if variation.starting_comment:
                visitor.visit_comment(variation.starting_comment)

            # Visit move.
            visitor.visit_move(board, variation.move)

            # Visit NAGs.
            for nag in sorted(variation.nags):
                visitor.visit_nag(nag)

            # Visit comment.
            if variation.comment:
                visitor.visit_comment(variation.comment)

            # Recursively append the next moves.
            board.push(variation.move)
            variation.accept(visitor, _board=board)
            board.pop()

            # End variation.
            visitor.end_variation()

        # The mainline is continued last.
        if self.variations:
            main_variation = self.variations[0]

            # Recursively append the next moves.
            board.push(main_variation.move)
            main_variation.accept(visitor, _board=board)
            board.pop()

        # Get the result if not called recursively.
        if _board is None:
            return visitor.result()

    def __str__(self):
        return self.accept(StringExporter(columns=None))


class Game(GameNode):
    """
    The root node of a game with extra information such as headers and the
    starting position.

    By default the following 7 headers are provided in an ordered dictionary:

    >>> game = chess.pgn.Game()
    >>> game.headers["Event"]
    '?'
    >>> game.headers["Site"]
    '?'
    >>> game.headers["Date"]
    '????.??.??'
    >>> game.headers["Round"]
    '?'
    >>> game.headers["White"]
    '?'
    >>> game.headers["Black"]
    '?'
    >>> game.headers["Result"]
    '*'

    Also has all the other properties and methods of
    :class:`~chess.pgn.GameNode`.
    """

    def __init__(self):
        super(Game, self).__init__()

        self.headers = collections.OrderedDict()
        self.headers["Event"] = "?"
        self.headers["Site"] = "?"
        self.headers["Date"] = "????.??.??"
        self.headers["Round"] = "?"
        self.headers["White"] = "?"
        self.headers["Black"] = "?"
        self.headers["Result"] = "*"

        self.errors = []

    def board(self, _cache=False):
        """
        Gets the starting position of the game.

        Unless the `SetUp` and `FEN` header tags are set this is the default
        starting position.
        """
        chess960 = self.headers.get("Variant", "").lower() == "chess960"

        if chess960 or "Variant" not in self.headers:
            VariantBoard = chess.Board
        else:
            from chess.variant import find_variant
            VariantBoard = find_variant(self.headers["Variant"])

        fen = self.headers.get("FEN") if self.headers.get("SetUp", "1") == "1" else None
        board = VariantBoard(fen or VariantBoard.starting_fen, chess960)
        board.chess960 = board.chess960 or board.has_chess960_castling_rights()
        return board

    def setup(self, board):
        """
        Setup a specific starting position. This sets (or resets) the *SetUp*,
        *FEN* and *Variant* header tags.
        """
        try:
            fen = board.fen()
        except AttributeError:
            board = chess.Board(board)
            board.chess960 = board.has_chess960_castling_rights()
            fen = board.fen()

        if fen == type(board).starting_fen:
            self.headers.pop("SetUp", None)
            self.headers.pop("FEN", None)
        else:
            self.headers["SetUp"] = "1"
            self.headers["FEN"] = fen

        if type(board).aliases[0] == "Standard" and board.chess960:
            self.headers["Variant"] = "Chess960"
        elif type(board).aliases[0] != "Standard":
            self.headers["Variant"] = type(board).aliases[0]
            self.headers["FEN"] = board.fen()
        else:
            self.headers.pop("Variant", None)

    def accept(self, visitor):
        """
        Traverses the game in PGN order using the given *visitor*. Returns
        the visitor result.
        """
        visitor.begin_game()

        visitor.begin_headers()
        for tagname, tagvalue in self.headers.items():
            visitor.visit_header(tagname, tagvalue)
        visitor.end_headers()

        if self.comment:
            visitor.visit_comment(self.comment)

        super(Game, self).accept(visitor, _board=self.board())

        visitor.visit_result(self.headers.get("Result", "*"))
        visitor.end_game()
        return visitor.result()

    @classmethod
    def from_board(cls, board):
        """Creates a game from the move stack of a :class:`~chess.Board()`."""
        # Undo all moves.
        switchyard = collections.deque()
        while board.move_stack:
            switchyard.append(board.pop())

        # Setup initial position.
        game = cls()
        game.setup(board)
        node = game

        # Replay all moves.
        while switchyard:
            move = switchyard.pop()
            node = node.add_variation(move)
            board.push(move)

        game.headers["Result"] = board.result()
        return game


class BaseVisitor(object):
    """
    Base class for visitors.

    Use with :func:`chess.pgn.Game.accept()` or
    :func:`chess.pgn.GameNode.accept()`.

    Methods are called in PGN order.
    """

    def begin_game(self):
        """Called at the start of a game."""
        pass

    def begin_headers(self):
        """Called at the start of the game headers."""
        pass

    def visit_header(self, tagname, tagvalue):
        """Called for each game header."""
        pass

    def end_headers(self):
        """Called at the end of the game headers."""
        pass

    def visit_move(self, board, move):
        """
        Called for each move.

        *board* is the board state before the move. The board state must be
        restored before the traversal continues.
        """
        pass

    def visit_comment(self, comment):
        """Called for each comment."""
        pass

    def visit_nag(self, nag):
        """Called for each NAG."""
        pass

    def begin_variation(self):
        """
        Called at the start of a new variation. It is not called for the
        mainline of the game.
        """
        pass

    def end_variation(self):
        """Concludes a variation."""
        pass

    def visit_result(self, result):
        """Called at the end of the game with the *Result*-header."""
        pass

    def end_game(self):
        """Called at the end of a game."""
        pass

    def result(self):
        """Called to get the result of the visitor. Defaults to ``True``."""
        return True

    def handle_error(self, error):
        """Called for errors encountered. Defaults to raising an exception."""
        raise error


class GameModelCreator(BaseVisitor):
    """
    Creates a game model. Default visitor for :func:`~chess.pgn.read_game()`.
    """

    def __init__(self):
        self.game = Game()

        self.variation_stack = collections.deque([self.game])
        self.starting_comment = ""
        self.in_variation = False

    def visit_header(self, tagname, tagvalue):
        self.game.headers[tagname] = tagvalue

    def visit_nag(self, nag):
        self.variation_stack[-1].nags.add(nag)

    def begin_variation(self):
        self.variation_stack.append(self.variation_stack[-1].parent)
        self.in_variation = False

    def end_variation(self):
        self.variation_stack.pop()

    def visit_result(self, result):
        if self.game.headers.get("Result", "*") == "*":
            self.game.headers["Result"] = result

    def visit_comment(self, comment):
        if self.in_variation or (not self.variation_stack[-1].parent and self.variation_stack[-1].is_end()):
            # Add as a comment for the current node if in the middle of
            # a variation. Add as a comment for the game, if the comment
            # starts before any move.
            new_comment = [self.variation_stack[-1].comment, comment]
            self.variation_stack[-1].comment = "\n".join(new_comment).strip()
        else:
            # Otherwise it is a starting comment.
            new_comment = [self.starting_comment, comment]
            self.starting_comment = "\n".join(new_comment).strip()

    def visit_move(self, board, move):
        self.variation_stack[-1] = self.variation_stack[-1].add_variation(move)
        self.variation_stack[-1].starting_comment = self.starting_comment
        self.starting_comment = ""
        self.in_variation = True

    def handle_error(self, error):
        """
        Populates :data:`chess.pgn.Game.errors` with encountered errors and
        logs them.
        """
        LOGGER.exception("error during pgn parsing")
        self.game.errors.append(error)

    def result(self):
        """
        Returns the visited :class:`~chess.pgn.Game()`.
        """
        return self.game


class StringExporter(BaseVisitor):
    """
    Allows exporting a game as a string.

    >>> exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
    >>> pgn_string = game.accept(exporter)

    Only *columns* characters are written per line. If *columns* is ``None``
    then the entire movetext will be on a single line. This does not affect
    header tags and comments.

    There will be no newlines at the end of the string.
    """

    def __init__(self, columns=80, headers=True, comments=True, variations=True):
        self.columns = columns
        self.headers = headers
        self.comments = comments
        self.variations = variations

        self.force_movenumber = True

        self.lines = []
        self.current_line = ""
        self.variation_depth = 0

    def flush_current_line(self):
        if self.current_line:
            self.lines.append(self.current_line.rstrip())
        self.current_line = ""

    def write_token(self, token):
        if self.columns is not None and self.columns - len(self.current_line) < len(token):
            self.flush_current_line()
        self.current_line += token

    def write_line(self, line=""):
        self.flush_current_line()
        self.lines.append(line.rstrip())

    def begin_game(self):
        self.after_variation = True

    def end_game(self):
        self.write_line()

    def visit_header(self, tagname, tagvalue):
        if self.headers:
            self.write_line("[{0} \"{1}\"]".format(tagname, tagvalue))

    def end_headers(self):
        if self.headers:
            self.write_line()

    def begin_variation(self):
        self.variation_depth += 1

        if self.variations:
            self.write_token("( ")
            self.force_movenumber = True

    def end_variation(self):
        self.variation_depth -= 1

        if self.variations:
            self.write_token(") ")
            self.force_movenumber = True

    def visit_comment(self, comment):
        if self.comments and (self.variations or not self.variation_depth):
            self.write_token("{ " + comment.replace("}", "").strip() + " } ")
            self.force_movenumber = True

    def visit_nag(self, nag):
        if self.comments and (self.variations or not self.variation_depth):
            self.write_token("$" + str(nag) + " ")

    def visit_move(self, board, move):
        if self.variations or not self.variation_depth:
            # Write the move number.
            if board.turn == chess.WHITE:
                self.write_token(str(board.fullmove_number) + ". ")
            elif self.force_movenumber:
                self.write_token(str(board.fullmove_number) + "... ")

            # Write the SAN.
            self.write_token(board.san(move) + " ")

            self.force_movenumber = False

    def visit_result(self, result):
        self.write_token(result + " ")

    def result(self):
        if self.current_line:
            return "\n".join(itertools.chain(self.lines, [self.current_line.rstrip()])).rstrip()
        else:
            return "\n".join(self.lines).rstrip()

    def __str__(self):
        return self.result()


class FileExporter(StringExporter):
    """
    Like a :class:`~chess.pgn.StringExporter`, but games are written directly
    to a text file.

    There will always be a blank line after each game. Handling encodings is up
    to the caller.

    >>> new_pgn = open("new.pgn", "w", encoding="utf-8")
    >>> exporter = chess.pgn.FileExporter(new_pgn)
    >>> game.accept(exporter)
    """

    def __init__(self, handle, columns=80, headers=True, comments=True, variations=True):
        super(FileExporter, self).__init__(columns=columns, headers=headers, comments=comments, variations=variations)
        self.handle = handle

    def flush_current_line(self):
        if self.current_line:
            self.handle.write(self.current_line.rstrip())
            self.handle.write("\n")
        self.current_line = ""

    def write_line(self, line=""):
        self.flush_current_line()
        self.handle.write(line.rstrip())
        self.handle.write("\n")

    def result(self):
        return None

    def __repr__(self):
        return "<FileExporter at {0}>".format(hex(id(self)))

    def __str__(self):
        return self.__repr__()


def read_game(handle, Visitor=GameModelCreator):
    """
    Reads a game from a file opened in text mode.

    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn")
    >>> first_game = chess.pgn.read_game(pgn)
    >>> second_game = chess.pgn.read_game(pgn)
    >>>
    >>> first_game.headers["Event"]
    'IBM Man-Machine, New York USA'

    By using text mode the parser does not need to handle encodings. It is the
    callers responsibility to open the file with the correct encoding.
    PGN files are ASCII or UTF-8 most of the time. So the following should
    cover most relevant cases (ASCII, UTF-8 without BOM, UTF-8 with BOM,
    UTF-8 with encoding errors).

    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn", encoding="utf-8-sig", errors="surrogateescape")

    Use `StringIO` to parse games from a string.

    >>> pgn_string = "1. e4 e5 2. Nf3 *"
    >>>
    >>> try:
    >>>     from StringIO import StringIO  # Python 2
    >>> except ImportError:
    >>>     from io import StringIO  # Python 3
    >>>
    >>> pgn = StringIO(pgn_string)
    >>> game = chess.pgn.read_game(pgn)

    The end of a game is determined by a completely blank line or the end of
    the file. (Of course blank lines in comments are possible.)

    According to the standard at least the usual 7 header tags are required
    for a valid game. This parser also handles games without any headers just
    fine.

    The parser is relatively forgiving when it comes to errors. It skips over
    tokens it can not parse. Any exceptions are logged.

    Returns the parsed game or ``None`` if the EOF is reached.
    """
    visitor = Visitor()

    dummy_game = Game()
    found_game = False
    found_content = False

    line = handle.readline()

    # Parse game headers.
    while line:
        # Skip empty lines and comments.
        if not line.strip() or line.strip().startswith("%"):
            line = handle.readline()
            continue

        if not found_game:
            visitor.begin_game()
            visitor.begin_headers()
            found_game = True

        # Read header tags.
        tag_match = TAG_REGEX.match(line)
        if tag_match:
            dummy_game.headers[tag_match.group(1)] = tag_match.group(2)
            visitor.visit_header(tag_match.group(1), tag_match.group(2))
        else:
            break

        line = handle.readline()

    if found_game:
        visitor.end_headers()

    # Get the next non-empty line.
    while not line.strip() and line:
        line = handle.readline()

    # Movetext parser state.
    try:
        board_stack = collections.deque([dummy_game.board()])
    except ValueError as error:
        visitor.handle_error(error)
        board_stack = collections.deque([chess.Board()])

    # Parse movetext.
    while line:
        read_next_line = True

        # An empty line is the end of a game.
        if not line.strip() and found_content:
            if found_game:
                visitor.end_game()
                return visitor.result()
            else:
                return

        for match in MOVETEXT_REGEX.finditer(line):
            token = match.group(0)

            if token.startswith("%"):
                # Ignore the rest of the line.
                line = handle.readline()
                continue

            if not found_game:
                found_game = True
                visitor.begin_game()

            if token.startswith("{"):
                # Consume until the end of the comment.
                line = token[1:]
                comment_lines = []
                while line and "}" not in line:
                    comment_lines.append(line.rstrip())
                    line = handle.readline()
                end_index = line.find("}")
                comment_lines.append(line[:end_index])
                if "}" in line:
                    line = line[end_index:]
                else:
                    line = ""

                visitor.visit_comment("\n".join(comment_lines).strip())

                # Continue with the current or the next line.
                if line:
                    read_next_line = False
                break
            elif token.startswith("$"):
                # Found a NAG.
                try:
                    nag = int(token[1:])
                except ValueError as error:
                    visitor.handle_error(error)
                else:
                    visitor.visit_nag(nag)
            elif token == "?":
                visitor.visit_nag(NAG_MISTAKE)
            elif token == "??":
                visitor.visit_nag(NAG_BLUNDER)
            elif token == "!":
                visitor.visit_nag(NAG_GOOD_MOVE)
            elif token == "!!":
                visitor.visit_nag(NAG_BRILLIANT_MOVE)
            elif token == "!?":
                visitor.visit_nag(NAG_SPECULATIVE_MOVE)
            elif token == "?!":
                visitor.visit_nag(NAG_DUBIOUS_MOVE)
            elif token == "(":
                if board_stack[-1].move_stack:
                    visitor.begin_variation()

                    board = board_stack[-1].copy()
                    board.pop()
                    board_stack.append(board)
            elif token == ")":
                # Found a close variation token. Always leave at least the
                # root node on the stack.
                if len(board_stack) > 1:
                    visitor.end_variation()
                    board_stack.pop()
            elif token in ["1-0", "0-1", "1/2-1/2", "*"] and len(board_stack) == 1:
                # Found a result token.
                found_content = True
                visitor.visit_result(token)
            else:
                # Found a SAN token.
                found_content = True

                # Replace zeros castling notation.
                if token == "0-0":
                    token = "O-O"
                elif token == "0-0-0":
                    token = "O-O-O"

                # Parse the SAN.
                try:
                    move = board_stack[-1].parse_san(token)
                except ValueError as error:
                    visitor.handle_error(error)
                else:
                    visitor.visit_move(board_stack[-1], move)
                    board_stack[-1].push(move)

        if read_next_line:
            line = handle.readline()

    if found_game:
        visitor.end_game()
        return visitor.result()


def scan_headers(handle):
    """
    Scan a PGN file opened in text mode for game offsets and headers.

    Yields a tuple for each game. The first element is the offset. The second
    element is an ordered dictionary of game headers.

    Since actually parsing many games from a big file is relatively expensive,
    this is a better way to look only for specific games and seek and parse
    them later.

    This example scans for the first game with Kasparov as the white player.

    >>> pgn = open("mega.pgn")
    >>> for offset, headers in chess.pgn.scan_headers(pgn):
    ...     if "Kasparov" in headers["White"]:
    ...         kasparov_offset = offset
    ...         break

    Then it can later be seeked an parsed.

    >>> pgn.seek(kasparov_offset)
    >>> game = chess.pgn.read_game(pgn)

    This also works nicely with generators, scanning lazily only when the next
    offset is required.

    >>> white_win_offsets = (offset for offset, headers in chess.pgn.scan_headers(pgn)
    ...                             if headers["Result"] == "1-0")
    >>> first_white_win = next(white_win_offsets)
    >>> second_white_win = next(white_win_offsets)

    :warning: Be careful when seeking a game in the file while more offsets are
        being generated.
    """
    in_comment = False

    game_headers = None
    game_pos = None

    last_pos = handle.tell()
    line = handle.readline()

    while line:
        # Skip single line comments.
        if line.startswith("%"):
            last_pos = handle.tell()
            line = handle.readline()
            continue

        # Reading a header tag. Parse it and add it to the current headers.
        if not in_comment and line.startswith("["):
            tag_match = TAG_REGEX.match(line)
            if tag_match:
                if game_pos is None:
                    game_headers = collections.OrderedDict()
                    game_headers["Event"] = "?"
                    game_headers["Site"] = "?"
                    game_headers["Date"] = "????.??.??"
                    game_headers["Round"] = "?"
                    game_headers["White"] = "?"
                    game_headers["Black"] = "?"
                    game_headers["Result"] = "*"

                    game_pos = last_pos

                game_headers[tag_match.group(1)] = tag_match.group(2)

                last_pos = handle.tell()
                line = handle.readline()
                continue

        # Reading movetext. Update parser state in_comment in order to skip
        # comments that look like header tags.
        if (not in_comment and "{" in line) or (in_comment and "}" in line):
            in_comment = line.rfind("{") > line.rfind("}")

        # Reading movetext. If there were headers, previously, those are now
        # complete and can be yielded.
        if game_pos is not None:
            yield game_pos, game_headers
            game_pos = None

        last_pos = handle.tell()
        line = handle.readline()

    # Yield the headers of the last game.
    if game_pos is not None:
        yield game_pos, game_headers


def scan_offsets(handle):
    """
    Scan a PGN file opened in text mode for game offsets.

    Yields the starting offsets of all the games, so that they can be seeked
    later. This is just like :func:`~chess.pgn.scan_headers()` but more
    efficient if you do not actually need the header information.

    The PGN standard requires each game to start with an *Event*-tag. So does
    this scanner.
    """
    in_comment = False

    last_pos = handle.tell()
    line = handle.readline()

    while line:
        if not in_comment and line.startswith("[Event \""):
            yield last_pos
        elif (not in_comment and "{" in line) or (in_comment and "}" in line):
            in_comment = line.rfind("{") > line.rfind("}")

        last_pos = handle.tell()
        line = handle.readline()



#
#   gv.py
#
#   This file is part of gchess
#
#   gchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   gchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with gchess.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

# Global variables

# gettext/locale translation domain
domain = "jcchess"

# set in run.py
installed = True

# set in gchess.py
verbose = False
verbose_uci = False
lastdir = ""
event = ""
gamedate = ""
gote = ""
sente = ""

# references to classes that have one instance
jcchess = None
gui = None
pieces = None
tc = None
board = None
usib = None
usiw = None
engine_manager = None
set_board_colours = None
load_save = None

# Make sure no gv global variables have been created elsewhere
# Only those above should be present

modulename = sys.modules[__name__]  # gchess.gv


def testnames():
    names = dir(modulename)
    for n in names:
        if n not in initial_namelist:
            print("unknown global variable in gv.py:  ",  n  )
initial_namelist = None     # don't remove this line
initial_namelist = dir()



#
#   pieces.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Rsvg
import os
import traceback
import chess

from . import gv

class Pieces:

    def __init__(self):
        self.pieceset = 1

    # called from jcchess.py
    def load_pieces(self, prefix):
        path = os.path.join(prefix, "images", "jcchess")
        p1 = self.load_pieceset(os.path.join(prefix, "images", "pieceset1"))
        p2 = self.load_pieceset(os.path.join(prefix, "images", "pieceset2"))
        self.piece_handles = (p1, p2)

    """
    def load_pieceset1(self):
        images = [
            "p", "n", "b", "r", "k", "q",
            "P", "N", "B", "R", "K", "Q",
            ]
        pieces = []

        for image in images:
            svg = chess.svg.piece(chess.Piece.from_symbol(image))
            svghandle = Rsvg.Handle.new_from_data(svg.encode())
            pieces.append(svghandle)
            #loader = GdkPixbuf.PixbufLoader()
            #loader.write(svg.encode())
            #loader.close()
            #pb = loader.get_pixbuf()
            #piece_pixbuf.append(pb)
        return pieces
    """

    def load_pieceset(self, pieces_dir):
        images = ["pawn", "knight", "bishop", "rook", "king", "queen"]
        pieces = []

        # Load pieces
        for side in ("B", "W"):
            for image in images:
                image = image + side + ".svg"
                path = os.path.join(pieces_dir, image)
                svghandle = Rsvg.Handle.new_from_file(path)
                pieces.append(svghandle)

        return pieces 

    def gethandle(self, piece): 
        # pieces contains the list of possible pieces
        pieces = [
            "None", "p", "n", "b", "r", "k", "q",
                  "P", "N", "B", "R", "K", "Q"]
        try:
            idx = pieces.index(piece)
        except ValueError as ve:
            traceback.print_exc()
            print("error piece not found, piece =", piece)
            print(len(piece))
        return self.piece_handles[self.pieceset][idx-1]

    def get_pieceset(self):
        return self.pieceset

    def set_pieceset(self, pieceset):
        if pieceset not in (0,1):
            print("attempt to set invalid pieceset:",pieceset)
            return
        self.pieceset = pieceset

    def show_pieces_dialog(self, gtkaction):
        dialog = Gtk.Dialog(
            _("Select Pieces"), gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OK, Gtk.ResponseType.OK))
        rb1 = Gtk.RadioButton.new_with_label(None, "Pieceset 1")
        dialog.vbox.pack_start(rb1, False, True, 5)
        rb2 = Gtk.RadioButton.new_with_label_from_widget(rb1, "Pieceset 2")
        dialog.vbox.pack_start(rb2, False, True, 5)
        dialog.show_all()
        dialog.set_default_response(Gtk.ResponseType.OK)
        if self.pieceset == 0:
            rb1.set_active(True)
        elif self.pieceset == 1:
            rb2.set_active(True)
        else:
            rb1.set_active()
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            if rb1.get_active():
                self.pieceset = 0
            elif rb2.get_active():
                self.pieceset = 1
        dialog.destroy()
        return



#
#   constants.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

NAME = "jcchess"
VERSION = "0.0.1"

# These values match those used in the engine
BLACK = 0
WHITE = 1
NEUTRAL = 2

# Drag and Drop
TARGET_TYPE_TEXT = 1180



#
#   engine_output.py - Display USI Engine Output Window
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Pango
import os

from . import gv


class Engine_Output:

    engine_output_ref = None

    def __init__(self):
        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "engine_output.glade")
        Engine_Output.engine_output_ref = self

        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("engine_output_window")
        self.tv = [Gtk.TextView(), Gtk.TextView()]
        self.tv[0] = self.builder.get_object("engine_output_textview1")
        self.tv[1] = self.builder.get_object("engine_output_textview2")
        self.tv[0].set_editable(False)
        self.tv[1].set_editable(False)

        tabs = Pango.TabArray.new(4, True)
        tabs.set_tab(0, Pango.TabAlign.LEFT, 40)
        tabs.set_tab(1, Pango.TabAlign.LEFT, 160)
        tabs.set_tab(2, Pango.TabAlign.LEFT, 230)
        tabs.set_tab(3, Pango.TabAlign.LEFT, 280)

        self.tv[0].set_tabs(tabs)
        self.tv[1].set_tabs(tabs)

        self.tb = [Gtk.TextBuffer(), Gtk.TextBuffer()]
        self.tb[0] = self.tv[0].get_buffer()
        self.tb[1] = self.tv[1].get_buffer()
        # self.tb[0].set_text("")
        # self.tb[1].set_text("")

        self.nps_lbl = [Gtk.Label(), Gtk.Label()]
        self.nps_lbl[0] = self.builder.get_object("engine_output_nodes_lbl1")
        self.nps_lbl[1] = self.builder.get_object("engine_output_nodes_lbl2")

        self.engine_name_lbl = [Gtk.Label(), Gtk.Label()]
        self.engine_name_lbl[0] = self.builder.get_object(
            "engine_output_engine_name_lbl1")
        self.engine_name_lbl[1] = self.builder.get_object(
            "engine_output_engine_name_lbl2")

        self.ponder_move_lbl = [Gtk.Label(), Gtk.Label()]
        self.ponder_move_lbl[0] = self.builder.get_object(
            "engine_output_ponder_move_lbl1")
        self.ponder_move_lbl[1] = self.builder.get_object(
            "engine_output_ponder_move_lbl2")

        self.currmove_lbl = [Gtk.Label(), Gtk.Label()]
        self.currmove_lbl[0] = self.builder.get_object(
            "engine_output_currmove_lbl1")
        self.currmove_lbl[1] = self.builder.get_object(
            "engine_output_currmove_lbl2")

        # self.window.show_all()

    # user has closed the window
    # just hide it
    def delete_event(self, widget, event):
        self.window.hide()
        return True  # do not propagate to other handlers

    def format_time(self, ztime):
        if ztime == "":
            return ztime
        try:
            ms = int(ztime)
        except:
            return ztime
        secs = int(ms / 1000)
        mins = 0
        if secs > 60:
            mins = int(secs / 60)
            secs = secs - mins * 60
        smins = str(mins)
        if mins < 10:
            smins = "0" + smins
        ssecs = str(secs)
        if secs < 10:
            ssecs = "0" + ssecs
        return smins + ":" + ssecs

    def add_to_log(self, side, engine_name, msg):
        # Write to either the black or white split pane
        if side == "b":
            idx = 1   # bottom pane for black
        else:
            idx = 0   # top pane for white

        msg = msg + "\n"

        ztime = ""
        nodes = ""
        depth = ""
        nps = ""
        pv = ""
        currmove = ""
        score = ""
        msg_lst = msg.split()
        for i in range(0, len(msg_lst)):
            if msg_lst[i] == "time":
                ztime = msg_lst[i + 1]
            elif msg_lst[i] == "nodes":
                nodes = msg_lst[i + 1]
            elif msg_lst[i] == "depth":
                depth = msg_lst[i + 1]
            elif msg_lst[i] == "nps":
                nps = msg_lst[i + 1]
            elif msg_lst[i] == "currmove":
                currmove = msg_lst[i + 1]
            elif msg_lst[i] == "score":
                score = msg_lst[i + 1]
                if score == "cp":
                    score = msg_lst[i + 2]
                elif score == "mate":
                    score = score + " " + msg_lst[i + 2]
            elif msg_lst[i] == "pv":
                pv_lst = msg_lst[i + 1:]
                for p in pv_lst:
                    pv = pv + p + "  "

        ztime = self.format_time(ztime)
        zmsg = (depth + "\t" + nodes + "\t" + ztime + "\t" +
                score + "\t" + pv + "\n")

        # insert at start of buffer
        start_iter = self.tb[idx].get_start_iter()

        if ztime != "" or nodes != "" or depth != "" or pv != "":
            self.tb[idx].insert(start_iter, zmsg)

        self.nps_lbl[idx].set_text("NPS: " + nps)
        """
        if nps != "":
            self.nps_lbl[idx].set_text("NPS: " + nps)
        else:
            try:
                nps_num = int(nodes) * 1000 / int(ztime)
                nps = str(nps_num)
                self.nps_lbl[idx].set_text("NPS: " + nps)
            except:
                pass
        """

        # self.nps_lbl[idx].set_text("NPS: 1200")

        if side == "b":
            s = _("Black") + ": "
        else:
            s = _("White") + ": "
        self.engine_name_lbl[idx].set_text(s + engine_name)

        if currmove != "":
            self.currmove_lbl[idx].set_text(_("Current Move") + ": " + currmove)

    def clear(self, side, engine_name):
        # Write to either the black or white split pane
        if side == "b":
            idx = 1   # bottom pane for black
        else:
            idx = 0   # top pane for white

        self.tb[idx].set_text("")

        if side == "b":
            s = _("Black") + ": "
        else:
            s = _("White") + ": "
        self.engine_name_lbl[idx].set_text(s + engine_name)

    def set_ponder_move(self, pondermove, side):
        if side == "b":
            self.ponder_move_lbl[1].set_text(_("Ponder") + ": " + pondermove)
        else:
            self.ponder_move_lbl[0].set_text(_("Ponder") + ": " + pondermove)

    def show_engine_output_window(self, b):
        self.window.present()


def get_ref():
    if Engine_Output.engine_output_ref is None:
        Engine_Output.engine_output_ref = Engine_Output()
    return Engine_Output.engine_output_ref



#
#   drag_and_drop.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import GObject
import cairo

from .constants import TARGET_TYPE_TEXT
from . import gv
from .board import SCALE

class Drag_And_Drop:

    drag_and_drop_ref = None

    #
    # user has begun to drag a piece
    #
    def drag_begin(self, widget, drag_context, data):

        self.dnd_data_received = False

        # get x,y co-ords of source square
        x, y = data

        if gv.verbose:
            print("in drag begin")
            print("data=", data)
            print("widget_name=", widget.get_name())
            print("source sq=", x, y)

        stm = gv.jcchess.get_side_to_move()

        # convert the x, y co-ords into the shogi representation
        # (e.g. 8, 6 is 1g)
        sq = gv.board.get_square_posn(x, y)

        self.src = sq
        if gv.verbose:
            print("source square: (x, y) = (", x, ",",  y, ") ", sq)
        self.src_x = x
        self.src_y = y

        # set the icon for the drag and drop to the piece that is being
        # dragged
        self.piece = gv.board.get_piece(x, y)
        # get width/height of board square
        a = gv.gui.get_event_box(x, y).get_allocation()
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, a.width, a.height)
        svghandle = gv.board.get_piece_handle(x, y)
        dim = svghandle.get_dimensions()

        cr = cairo.Context(surface)
        sfw = (a.width * 1.0 / dim.width) * SCALE
        sfh = (a.height * 1.0 / dim.height) * SCALE
        cr.scale(sfw, sfh)

        svghandle.render_cairo(cr)
        # set mouse pointer to be in middle of drag icon
        surface.set_device_offset(-a.width/2, -a.height/2)
        Gtk.drag_set_icon_surface(drag_context, surface)

        # clear the square where the piece is being moved from
        gv.board.set_square_as_unoccupied(x, y)

    def sendCallback(self, widget, context, selection, targetType, eventTime):
        if targetType == TARGET_TYPE_TEXT:
            sel = "jcchess"
            selection.set_text(sel, 8)

    def receiveCallback(self, widget, context, x, y, selection, targetType,
                        time, data):
        if gv.verbose:
            print("in receive callback")
            print("x=", x)
            print("y=", y)
            print("selection.data=", selection.get_text())
            print("targetType=", targetType)
            print("time=", time)
            print("data=", data)

        self.dnd_data_received = True

        # get x,y co-ords of dest square
        x, y = data

        # convert the x, y co-ords into the shogi representation
        # (e.g. 8, 6 is 1g)
        sq = gv.board.get_square_posn(x, y)

        # set destination square
        dst = sq
        if gv.verbose:
            print("dst =", dst)

        move = gv.jcchess.get_move(self.piece, self.src, dst, self.src_x,
                                  self.src_y, x, y)
        if gv.verbose:
            print("move=", move)
            print()

        # if drag and drop failed then reinstate the piece where it
        # was dragged from
        if move is None:
            gv.board.update()
            return

        # display the move
        GObject.idle_add(gv.jcchess.human_move, move)

    # if drag and drop failed then reinstate the piece where it
    # was dragged from
    def drag_end(self, widget, drag_context):
        # if receiveCallback function not entered then restore board
        # to before the drag started
        if not self.dnd_data_received:
            gv.board.update()
            return


def get_ref():
    if Drag_And_Drop.drag_and_drop_ref is None:
        Drag_And_Drop.drag_and_drop_ref = Drag_And_Drop()
    return Drag_And_Drop.drag_and_drop_ref



#
#   time_control.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
import time

from . import gv
from .constants import WHITE, BLACK


class Time_Control:

    def __init__(self):

        # time control defaults
        self.type = 2

        # byoyomi
        self.byo_hours = 0
        self.byo_minutes = 45
        self.byo_byoyomi = 10

        # classical
        self.init_cls_settings()

        # incremental
        self.inc_hours = 0
        self.inc_minutes = 5
        self.inc_bonus = 0

        # fixed time per move
        self.ftpm_seconds = 5

        # search depth
        self.dpth_depth = 4

        # search nodes
        self.nodes_nodes = 50000

    def init_cls_settings(self):
        self.cls_max_sessions = 3
        self.cls_settings = []
        for i in range(0, self.cls_max_sessions):
            moves_to_go = 40
            minutes = 5
            if i == 0:
                repeat_times = 1
            else:
                repeat_times = 0
            self.cls_settings.append((moves_to_go, minutes, repeat_times))

    #
    # called to initialise the clocks at the start of a new game
    #
    def reset_clock(self):
        # fields for go command
        # byoyomi
        if self.type == 0:
            self.wtime = (
                self.byo_hours * 60 * 60 + self.byo_minutes * 60) * 1000
            self.btime = self.wtime
        # classical
        elif self.type == 1:
            # fields for go command
            self.cls_bsession = 0
            self.cls_wsession = 0
            # set session to first with non-zero repeat value
            for i in range(0, self.cls_max_sessions):
                moves_to_go, mins, rep = self.cls_settings[i]
                if rep != 0:
                    self.cls_bsession = i
                    self.cls_wsession = i
                    break

            (moves_to_go, minutes, repeat_times) = self.cls_settings[
                                                        self.cls_bsession]
            self.wtime = int((minutes * 60) * 1000)
            self.btime = self.wtime
            self.wmoves_to_go = int(moves_to_go)
            self.bmoves_to_go = int(moves_to_go)
            self.wrepeat = repeat_times
            self.brepeat = repeat_times
        # incremental
        elif self.type == 2:
            self.wtime = (
                self.inc_hours * 60 * 60 + self.inc_minutes * 60) * 1000
            self.btime = self.wtime
        # fixed time per move
        elif self.type == 3:
            self.wtime = int(self.ftpm_seconds) * 1000
            self.btime = self.wtime

        self.set_toolbar_time_control(self.type, 0, WHITE)
        self.set_toolbar_time_control(self.type, 0, BLACK)

    #
    # get clock settings so they can be saved when quitting jcchess
    #
    def get_clock_settings(self):
        return (
            self.type, self.byo_hours, self.byo_minutes, self.byo_byoyomi,
            self.inc_hours, self.inc_minutes, self.inc_bonus,
            self.cls_settings, self.ftpm_seconds, self.dpth_depth,
            self.nodes_nodes)

    #
    # restore the clock settings to the values from the previous game when
    # starting up jcchess
    #
    def restore_clock_settings(self, clock_settings):
        (
            self.type, self.byo_hours, self.byo_minutes, self.byo_byoyomi,
            self.inc_hours, self.inc_minutes, self.inc_bonus, cls_settings,
            self.ftpm_seconds, self.dpth_depth, self.nodes_nodes
        ) = clock_settings

        # if classical settings can't be restored then init them to default
        # values
        if len(cls_settings) == 0:
            self.init_cls_settings()
        else:
            self.cls_settings = cls_settings

        # wtime/btime for go command
        # byoyomi
        if self.type == 0:
            self.wtime = (
                self.byo_hours * 60 * 60 + self.byo_minutes * 60) * 1000
            self.btime = self.wtime
        # incremental
        elif self.type == 2:
            self.wtime = (
                self.inc_hours * 60 * 60 + self.inc_minutes * 60) * 1000
            self.btime = self.wtime
        else:
            self.wtime = (
                self.byo_hours * 60 * 60 + self.byo_minutes * 60) * 1000
            self.btime = self.wtime

    #
    # gui for changing time control settings
    #
    def time_control(self, b):
        dialog = Gtk.Dialog(
            "Time Control", gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK,
             Gtk.ResponseType.OK))
        dialog.set_title("Time Control")
        dialog.vbox.set_spacing(20)

        # list of time control types
        combobox = Gtk.ComboBoxText()
        #combobox.append_text("Byoyomi")
        combobox.append_text("Classical")
        combobox.append_text("Incremental")
        combobox.append_text("Fixed Time Per Move")
        combobox.append_text("Fixed Search Depth")
        combobox.append_text("Infinite Search")
        combobox.append_text("Nodes")
        combobox.set_active(self.type-1)

        al = Gtk.Alignment.new(xalign=0.0, yalign=1.0, xscale=0.0, yscale=0.5)
        # top, bottom, left, right
        al.set_padding(9, 0, 9, 9)
        al.add(combobox)
        dialog.vbox.pack_start(al, False, True, 5)
        self.combobox = combobox

        dialog.connect("draw", self.dialog_expose_event)

        #
        # settings for the byoyomi time control type
        #
        #   byoyomi
        #       e.g. 60 mins available time plus 10 seconds byoyomi
        #       go btime 3600000 wtime 3600000 byoyomi 10000
        #

        # list of time control vboxes. 1 per time control type
        self.tcvb = [Gtk.VBox(False, 0) for x in range(8)]

        byo_frame1 = Gtk.Frame.new("Available Time")
        vb = Gtk.VBox(False, 0)
        byo_frame1.add(vb)

        # available time - hours
        minimum = 0
        maximum = 10
        default = self.byo_hours
        byo_adj_hours = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(byo_adj_hours, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        hb = Gtk.HBox(False, 0)
        hb.pack_start(Gtk.Label("Hours:"), False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        # available time - minutes
        minimum = 0
        maximum = 59
        default = self.byo_minutes
        byo_adj_mins = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(byo_adj_mins, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        lbl = Gtk.Label(label="Minutes :")
        hb = Gtk.HBox(False, 0)
        hb.pack_start(lbl, False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        # byoyomi - seconds
        byo_frame2 = Gtk.Frame.new("Byoyomi")
        minimum = 0
        maximum = 60
        default = self.byo_byoyomi
        byo_adj_byoyomi = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(byo_adj_byoyomi, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        hb = Gtk.HBox(False, 0)
        hb.pack_start(Gtk.Label("Seconds:"), False, False, 0)
        hb.pack_start(al, True, True, 10)
        byo_frame2.add(hb)

        self.tcvb[0].pack_start(byo_frame1, False, False, 0)
        self.tcvb[0].pack_start(byo_frame2, False, False, 0)
        self.tcvb[0].set_spacing(20)

        #
        # settings for the classical time control type
        #
        #   classical
        #       e.g. 40 moves in 5 minutes
        #       go wtime 300000 btime 300000 movestogo 40
        #

        cls_frame1 = Gtk.Frame()
        vb = Gtk.VBox(False, 0)
        cls_frame1.add(vb)

        adj_cls_settings = []

        for i in range(0, self.cls_max_sessions):

            (moves_to_go, minutes, repeat_times) = self.cls_settings[i]

            # session
            hb = Gtk.HBox(False, 0)
            if i != 0:
                vb.pack_start(Gtk.HSeparator(), False, True, 0)
            hb.pack_start(Gtk.Label("#" + str(i + 1)), False, False, 0)
            # if i != 0:
            #   hb.pack_start(Gtk.CheckButton(, True, True, 0), True, True, 10)
            vb.pack_start(hb, False, True, 0)

            # moves
            minimum = 1
            maximum = 200
            default = moves_to_go
            # default = 40
            # if i == 0:
            #    default = 40
            # else:
            #    default = 0
            adj_moves_to_go = Gtk.Adjustment(
                value=default, lower=minimum, upper=maximum,
                step_increment=1, page_increment=5, page_size=0)
            spinner = Gtk.SpinButton.new(adj_moves_to_go, 1.0, 0)
            al = Gtk.Alignment.new(
                xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
            al.add(spinner)
            hb = Gtk.HBox(False, 0)
            hb.pack_start(Gtk.Label("Moves:"), False, False, 0)
            hb.pack_start(al, True, True, 10)
            vb.pack_start(hb, False, True, 0)

            # minutes
            minimum = 0
            maximum = 500
            # default = 5
            default = minutes
            # if i == 0:
            #    default = 5
            # else:
            #    default = 0
            adj_cls_mins = Gtk.Adjustment(
                value=default, lower=minimum, upper=maximum,
                step_increment=1, page_increment=5, page_size=0)
            spinner = Gtk.SpinButton.new(adj_cls_mins, 1.0, 0)
            al = Gtk.Alignment.new(
                xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
            al.add(spinner)
            lbl = Gtk.Label(label="Minutes :")
            hb = Gtk.HBox(False, 0)
            hb.pack_start(lbl, False, False, 0)
            hb.pack_start(al, True, True, 10)
            vb.pack_start(hb, False, True, 0)

            # repeating
            minimum = 0
            maximum = 9
            default = repeat_times
            # if i == 0:
            #    default = 1
            # else:
            #    default = 0
            adj_cls_repeat = Gtk.Adjustment(
                value=default, lower=minimum, upper=maximum,
                step_increment=1, page_increment=5, page_size=0)
            spinner = Gtk.SpinButton.new(adj_cls_repeat, 1.0, 0)
            al = Gtk.Alignment.new(
                xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
            al.add(spinner)
            lbl = Gtk.Label(label="Count :")
            hb = Gtk.HBox(False, 0)
            hb.pack_start(lbl, False, False, 0)
            hb.pack_start(al, True, True, 10)
            vb.pack_start(hb, False, True, 0)

            # enabled
            # lbl = Gtk.Label(label="Enabled :")
            # hb = Gtk.HBox(False, 0)
            # hb.pack_start(lbl, False, False, 0)
            # hb.pack_start(Gtk.CheckButton(, True, True, 0), True, True, 10)
            # vb.pack_start(hb, False, True, 0)

            adj_cls_settings.append((
                adj_moves_to_go, adj_cls_mins, adj_cls_repeat))

        self.tcvb[1].pack_start(cls_frame1, False, False, 0)

        # Blitz (incremental)
        # e.g. 2 mins 0 seconds for whole game, 6 seconds 0 milliseconds bonus
        # per move
        # base is for whole game, bonus will be given for every move made
        # go wtime 126000 btime 120000 winc 6000 binc 6000
        # go wtime 130591 btime 118314 winc 6000 binc 6000
        # go wtime 135329 btime 118947 winc 6000 binc 6000

        inc_frame1 = Gtk.Frame.new("Time")
        vb = Gtk.VBox(False, 0)
        inc_frame1.add(vb)

        # available time - hours
        minimum = 0
        maximum = 10
        default = self.inc_hours
        inc_adj_hours = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(inc_adj_hours, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        hb = Gtk.HBox(False, 0)
        hb.pack_start(Gtk.Label("Hours:"), False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        # available time - minutes
        minimum = 0
        maximum = 59
        default = self.inc_minutes
        inc_adj_mins = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(inc_adj_mins, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        lbl = Gtk.Label(label="Minutes :")
        hb = Gtk.HBox(False, 0)
        hb.pack_start(lbl, False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        # bonus time per move - seconds
        inc_frame2 = Gtk.Frame.new("Bonus Time per Move")
        minimum = 0
        maximum = 60
        default = self.inc_bonus
        inc_adj_bonus = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(inc_adj_bonus, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        hb = Gtk.HBox(False, 0)
        hb.pack_start(Gtk.Label("Seconds:"), False, False, 0)
        hb.pack_start(al, True, True, 10)
        inc_frame2.add(hb)

        self.tcvb[2].pack_start(inc_frame1, False, False, 0)
        self.tcvb[2].pack_start(inc_frame2, False, False, 0)
        self.tcvb[2].set_spacing(20)

        #
        # settings for the fixed time per move time control type
        #
        #   fixed time per move
        #       e.g. 6 seconds per move
        #       go movetime 6000

        # ftpm_frame1 = Gtk.Frame("Fixed Time Per Move")
        ftpm_frame1 = Gtk.Frame()
        ftpm_frame1.set_shadow_type(Gtk.ShadowType.NONE)
        vb = Gtk.VBox(False, 0)
        ftpm_frame1.add(vb)

        # seconds per move
        minimum = 0
        maximum = 10000
        default = self.ftpm_seconds
        adj_ftpm_secs = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(adj_ftpm_secs, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        lbl = Gtk.Label(label="Seconds :")
        hb = Gtk.HBox(False, 0)
        hb.pack_start(lbl, False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        self.tcvb[3].pack_start(ftpm_frame1, False, False, 0)

        #
        # settings for the fixed search depth time control type
        #
        #   fixed search depth
        #       e.g.
        #       go depth 8
        dpth_frame1 = Gtk.Frame()
        dpth_frame1.set_shadow_type(Gtk.ShadowType.NONE)
        vb = Gtk.VBox(False, 0)
        dpth_frame1.add(vb)

        # depth
        minimum = 0
        maximum = 999
        default = self.dpth_depth
        adj_dpth_depth = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(adj_dpth_depth, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        lbl = Gtk.Label(label="Search Depth :")
        hb = Gtk.HBox(False, 0)
        hb.pack_start(lbl, False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        self.tcvb[4].pack_start(dpth_frame1, False, False, 0)

        # calculate to engines maximum search depth
        # (sends back bestmove if it gets the stop command)
        # go infinite

        # there are no cutomisable options for go infinite
        self.tcvb[5].pack_start(
            Gtk.Label("No customisable options"), True, True, 0)

        #
        # settings for the fixed no. of nodes time control type
        #
        #   fixed no. of nodes
        #       e.g. search for 1 million nodes
        #       go nodes 1000000
        nodes_frame1 = Gtk.Frame()
        nodes_frame1.set_shadow_type(Gtk.ShadowType.NONE)
        vb = Gtk.VBox(False, 0)
        nodes_frame1.add(vb)

        # nodes
        minimum = 1
        maximum = 2000000000
        default = self.nodes_nodes
        adj_nodes_nodes = Gtk.Adjustment(
            value=default, lower=minimum, upper=maximum,
            step_increment=1, page_increment=5, page_size=0)
        spinner = Gtk.SpinButton.new(adj_nodes_nodes, 1.0, 0)
        al = Gtk.Alignment.new(xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
        al.add(spinner)
        lbl = Gtk.Label(label="No. of Nodes to Search :")
        hb = Gtk.HBox(False, 0)
        hb.pack_start(lbl, False, False, 0)
        hb.pack_start(al, True, True, 10)
        vb.pack_start(hb, False, True, 0)

        self.tcvb[6].pack_start(nodes_frame1, False, False, 0)

        # mate search
        # e.g. look for mate in 5 moves
        # go mate 5

        self.al = Gtk.Alignment.new(
            xalign=0.5, yalign=0.5, xscale=1.0, yscale=0.0)
        # top, bottom, left, right
        self.al.set_padding(0, 0, 9, 9)
        self.al.add(self.tcvb[self.type])
        dialog.vbox.pack_start(self.al, False, True, 0)

        combobox.connect("changed", self.tc_method_changed, dialog)

        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.show_all()

        self.set_frame_visibility(self.type)

        while True:
            response = dialog.run()
            if response != Gtk.ResponseType.OK:
                # cancelled - exit loop
                break
            # OK pressed - validate input"
            # print "ok pressed"

            self.type = combobox.get_active() + 1
            # byoyomi
            if self.type == 0:
                byo_hours = int(byo_adj_hours.get_value())
                byo_minutes = int(byo_adj_mins.get_value())
                byo_byoyomi = int(byo_adj_byoyomi.get_value())
                if byo_hours == 0 and byo_minutes == 0 and byo_byoyomi == 0:
                    gv.gui.info_box("Time fields cannot all be zero!")
                else:
                    # input ok - exit loop
                    self.byo_hours = byo_hours
                    self.byo_minutes = byo_minutes
                    self.byo_byoyomi = byo_byoyomi

                    self.reset_clock()

                    self.set_toolbar_time_control(self.type, 0, WHITE)
                    self.set_toolbar_time_control(self.type, 0, BLACK)
                    break
            # classical
            elif self.type == 1:
                self.cls_settings = []
                rep_zero = True
                for i in range(0, self.cls_max_sessions):
                    moves_to_go, mins, rep = adj_cls_settings[i]
                    if rep.get_value() != 0:
                        rep_zero = False
                if rep_zero:
                    gv.gui.info_box("Count fields cannot all be zero!")
                else:
                    for i in range(0, self.cls_max_sessions):
                        moves_to_go, mins, rep = adj_cls_settings[i]
                        self.cls_settings.append((
                            moves_to_go.get_value(), mins.get_value(),
                            rep.get_value()))

                    # fields for go command
                    self.reset_clock()

                    self.set_toolbar_time_control(self.type, 0, WHITE)
                    self.set_toolbar_time_control(self.type, 0, BLACK)
                    break
            # incremental
            if self.type == 2:
                inc_hours = int(inc_adj_hours.get_value())
                inc_minutes = int(inc_adj_mins.get_value())
                inc_bonus = int(inc_adj_bonus.get_value())
                if inc_hours == 0 and inc_minutes == 0 and inc_bonus == 0:
                    gv.gui.info_box(
                        "Incremental Time fields cannot all be zero!")
                else:
                    # input ok - exit loop
                    self.inc_hours = inc_hours
                    self.inc_minutes = inc_minutes
                    self.inc_bonus = inc_bonus

                    self.reset_clock()

                    self.set_toolbar_time_control(self.type, 0, WHITE)
                    self.set_toolbar_time_control(self.type, 0, BLACK)
                    break
            # fixed time per move
            elif self.type == 3:
                self.ftpm_seconds = int(adj_ftpm_secs.get_value())

                # fields for go command
                self.reset_clock()

                self.set_toolbar_time_control(self.type, 0, WHITE)
                self.set_toolbar_time_control(self.type, 0, BLACK)
                break
            # fixed search depth
            elif self.type == 4:
                self.dpth_depth = int(adj_dpth_depth.get_value())

                # fields for go command
                self.reset_clock()

                self.set_toolbar_time_control(self.type, 0, WHITE)
                self.set_toolbar_time_control(self.type, 0, BLACK)
                break
            # infinite search
            elif self.type == 5:
                break
            # fixed number of nodes search
            elif self.type == 6:
                self.nodes_nodes = int(adj_nodes_nodes.get_value())

                # fields for go command
                self.reset_clock()

                self.set_toolbar_time_control(self.type, 0, WHITE)
                self.set_toolbar_time_control(self.type, 0, BLACK)
                break

        dialog.destroy()

    def tc_method_changed(self, b, dialog):
        tc_type = b.get_active() + 1

        self.set_frame_visibility(tc_type)

        chd = self.al.get_children()
        for c in chd:
            self.al.remove(c)
        self.al.add(self.tcvb[tc_type])
        self.tcvb[tc_type].show()

        dialog.size_allocate(self.area)
        dialog.check_resize()
        dialog.show_all()

    def set_frame_visibility(self, tc_type):
        for v in self.tcvb:
            v.hide()
        self.tcvb[tc_type].show()

    def dialog_expose_event(self, widget, context):
        # self.area = event.area
        a = widget.get_allocation()
        self.area = a
        # self.area = a.x, a.y, a.width, a.height
        # print self.area

    #
    # called before each move
    #
    def start_clock(self, stm):
        # print "starting clock in tc.py"
        self.clock_start_time = time.time()
        self.clock_stm = stm
        if self.type == 3:
            if stm == WHITE:
                self.wtime = self.ftpm_seconds * 1000
            else:
                self.btime = self.ftpm_seconds * 1000
        self.set_toolbar_time_control(self.type, 0, self.clock_stm)

    #
    # called when the red stop button is clicked
    #
    def stop_clock(self):
        elapsed_time = (time.time() - self.clock_start_time) * 1000
        if self.clock_stm == WHITE:
            self.wtime = int(self.wtime - elapsed_time)
            if self.wtime < 0:
                self.wtime = 0
        else:
            self.btime = int(self.btime - elapsed_time)
            if self.btime < 0:
                self.btime = 0

        # print "calling set_toolbar_time_control from stop_clock"
        self.set_toolbar_time_control(self.type, 0, self.clock_stm)

    #
    # This routine is called after a move from usi.py (if USI engine) or
    # jcchess.py
    #
    def update_clock(self):
        # set btime/wtime
        elapsed_time = (time.time() - self.clock_start_time) * 1000
        if self.clock_stm == WHITE:
            self.wtime = int(self.wtime - elapsed_time)
            if self.wtime < 0:
                self.wtime = 0
        else:
            self.btime = int(self.btime - elapsed_time)
            if self.btime < 0:
                self.btime = 0

        # set moves to go for classical TC
        if self.type == 1:
            if self.clock_stm == WHITE:
                self.wmoves_to_go -= 1
                if self.wmoves_to_go < 1:
                    self.wrepeat -= 1
                    while self.wrepeat < 1:
                        self.cls_wsession += 1
                        if self.cls_wsession >= self.cls_max_sessions:
                            self.cls_wsession = 0
                        self.wrepeat = self.cls_settings[self.cls_wsession][2]

                    (moves_to_go, minutes, repeat_times) = self.cls_settings[
                        self.cls_wsession]
                    rem_time = self.wtime
                    self.wtime = int((minutes * 60) * 1000) + rem_time
                    self.wmoves_to_go = int(moves_to_go)
            else:
                self.bmoves_to_go -= 1
                if self.bmoves_to_go < 1:
                    self.brepeat -= 1
                    while self.brepeat < 1:
                        self.cls_bsession += 1
                        if self.cls_bsession >= self.cls_max_sessions:
                            self.cls_bsession = 0
                        self.brepeat = self.cls_settings[self.cls_bsession][2]

                    (moves_to_go, minutes, repeat_times) = self.cls_settings[
                        self.cls_bsession]
                    rem_time = self.btime
                    self.btime = int((minutes * 60) * 1000) + rem_time
                    self.bmoves_to_go = int(moves_to_go)
        # incremental
        elif self.type == 2:
            if self.clock_stm == WHITE:
                self.wtime += self.inc_bonus * 1000
            else:
                self.btime += self.inc_bonus * 1000

        if self.type != 0:
            # print "calling set_toolbar_time_control from update_clock"
            self.set_toolbar_time_control(self.type, 0, self.clock_stm)

        # set stm to None
        # This prevents show_time from putting erroneous values in for the side
        # that has just moved
        # It will get set to the next side to move when start_clock is called
        # from usi.py
        self.clock_stm = None

    #
    # timer to countdown the clock on the screen during a move
    # called once per second while the player is thinking
    #
    def show_time(self):
        self.timer_active = True
        # don't refresh clock if stm is not set
        if self.clock_stm is None:
            return True
        if gv.jcchess.get_stopped():
            self.timer_active = False
            return False
        elapsed_time = int((time.time() - self.clock_start_time) * 1000)

        # print "calling set_toolbar_time_control from show_time"
        self.set_toolbar_time_control(self.type, elapsed_time, self.clock_stm)

        return True

    def get_go_command(self, stm):
        # byoyomi
        if self.type == 0:
            # times in milliseconds
            btime = self.btime
            wtime = self.wtime
            byoyomi = self.byo_byoyomi * 1000
            cmd = ("go btime " + str(btime) + " wtime " + str(wtime) +
                   " byoyomi " + str(byoyomi))
        # classical (e.g. 40 moves in 5 minutes)
        elif self.type == 1:
            if stm == WHITE:
                cmd = ("go btime " + str(self.btime) + " wtime " +
                       str(self.wtime) + " movestogo " +
                       str(self.wmoves_to_go))
            else:
                cmd = ("go btime " + str(self.btime) + " wtime " +
                       str(self.wtime) + " movestogo " +
                       str(self.bmoves_to_go))
        # incremental
        elif self.type == 2:
            # times in milliseconds
            btime = self.btime
            wtime = self.wtime
            bonus = self.inc_bonus * 1000
            cmd = ("go btime " + str(btime) + " wtime " + str(wtime) +
                   " binc " + str(bonus) + " winc " + str(bonus))
        # fixed time per move
        elif self.type == 3:
            cmd = "go movetime " + str(self.ftpm_seconds * 1000)
        elif self.type == 4:
            cmd = "go depth " + str(self.dpth_depth)
        elif self.type == 5:
            cmd = "go infinite"
        elif self.type == 6:
            cmd = "go nodes " + str(self.nodes_nodes)
        else:
            # times in milliseconds
            btime = self.btime
            wtime = self.wtime
            byoyomi = self.byo_byoyomi * 1000
            cmd = ("go btime " + str(btime) + " wtime " + str(wtime) +
                   " byoyomi " + str(byoyomi))

        return cmd

    # set the time limit/level when using the builtin jcchess engine
    def set_jcchess_time_limit(self, stm):

        #
        # set engine depth/nodes/time to maximum
        #
        engine.depth(39)
        engine.nodes(0)
        engine.command("level 0 59:59")

        # byoyomi
        if self.type == 0:
            if stm == WHITE:
                gs_movetime = self.wtime
            else:
                gs_movetime = self.btime
            gs_byoyomi = self.byo_byoyomi * 1000
            if gs_movetime > 0:
                gs_movetime = int(gs_movetime / 80)
            elif gs_byoyomi > 0:
                gs_movetime = gs_byoyomi
            else:
                # no time left - set to 1 second
                gs_movetime = 1000

            # if less than 1 second set to 1 second
            if gs_movetime < 1000:
                gs_movetime = 1000

            gs_secs = int(gs_movetime / 1000)
            gs_mins = int(gs_secs / 60)
            if gs_mins > 59:
                gs_mins = 59
                gs_secs = 59
            elif gs_mins > 0:
                gs_secs = gs_secs % (gs_mins * 60)

            gs_mins = int(gs_mins)
            gs_secs = int(gs_secs)

            gs_mins_str = str(gs_mins)
            gs_secs_str = str(gs_secs)
            if gs_mins < 10:
                gs_mins_str = "0" + gs_mins_str
            if gs_secs < 10:
                gs_secs_str = "0" + gs_secs_str

            gs_level = gs_mins_str + ":" + gs_secs_str

            if gv.verbose:
                print(("using byoyomi TC - wtime:", self.wtime, ", btime:",
                      self.btime, ", byoyomi:", self.byo_byoyomi * 1000,
                      ", stm:", stm, ", time for move (ms):", gs_movetime))
            command = "level 0 " + gs_level
            # print "jcchess time limit:",command

            # e.g.
            # time limit of 10 seconds per move
            # level 0 00:10
            #
            # time limit of 10 moves in 2 minutes
            # level 10 2
            engine.command(command)
        # classical
        elif self.type == 1:
            if stm == WHITE:
                gs_movetime = self.wtime
            else:
                gs_movetime = self.btime
            # if less than 1 second set to 1 second
            if gs_movetime < 1000:
                gs_movetime = 1000

            gs_secs = int(gs_movetime / 1000)
            gs_mins = int(gs_secs / 60)
            if gs_mins > 0:
                gs_secs = gs_secs % (gs_mins * 60)

            gs_mins = int(gs_mins)
            gs_secs = int(gs_secs)

            gs_mins_str = str(gs_mins)
            gs_secs_str = str(gs_secs)
            if gs_mins < 10:
                gs_mins_str = "0" + gs_mins_str
            if gs_secs < 10:
                gs_secs_str = "0" + gs_secs_str

            gs_level = gs_mins_str + ":" + gs_secs_str
            if gv.verbose:
                print(("using classical TC - wtime:", self.wtime, ", btime:",
                      self.btime, ", movestogo:", str(self.wmoves_to_go),
                      ", stm:", stm, ", time for move (ms):", gs_movetime))
                print(("                   - ", str(self.wmoves_to_go),
                      " moves in ", gs_mins_str, "minutes", gs_secs_str,
                      " seconds"))

            if self.wmoves_to_go == 1:
                command = "level 0 " + gs_level
            else:
                command = "level " + str(self.wmoves_to_go) + " " + gs_level
            engine.command(command)
        # incremental
        elif self.type == 2:
            if stm == WHITE:
                gs_movetime = self.wtime
            else:
                gs_movetime = self.btime

            if gs_movetime > 0:
                gs_movetime = int(gs_movetime / 80)
            else:
                # no time left - set to 1 second
                gs_movetime = 1000

            gs_secs = int(gs_movetime / 1000)
            gs_mins = int(gs_secs / 60)
            if gs_mins > 0:
                gs_secs = gs_secs % (gs_mins * 60)

            gs_mins = int(gs_mins)
            gs_secs = int(gs_secs)

            if gs_mins <= 0 and gs_secs < 1:
                gs_secs = 1

            gs_mins_str = str(gs_mins)
            gs_secs_str = str(gs_secs)
            if gs_mins < 10:
                gs_mins_str = "0" + gs_mins_str
            if gs_secs < 10:
                gs_secs_str = "0" + gs_secs_str

            gs_level = gs_mins_str + ":" + gs_secs_str
            if gv.verbose:
                print(("using incremental TC - wtime:", self.wtime, ", btime:",
                      self.btime, ", winc/binc:", str(self.inc_bonus * 1000),
                      ", stm:", stm, ", time for move (ms):", gs_movetime))

            command = "level 0 " + gs_level
            engine.command(command)
        # fixed time per move
        elif self.type == 3:
            gs_movetime = self.ftpm_seconds * 1000

            gs_secs = int(gs_movetime / 1000)
            gs_mins = int(gs_secs / 60)
            if gs_mins > 0:
                gs_secs = gs_secs % (gs_mins * 60)

            gs_mins = int(gs_mins)
            gs_secs = int(gs_secs)

            if gs_mins <= 0 and gs_secs < 1:
                gs_secs = 1

            gs_mins_str = str(gs_mins)
            gs_secs_str = str(gs_secs)
            if gs_mins < 10:
                gs_mins_str = "0" + gs_mins_str
            if gs_secs < 10:
                gs_secs_str = "0" + gs_secs_str

            gs_level = gs_mins_str + ":" + gs_secs_str
            if gv.verbose:
                print(("using fixed time per move TC - wtime:", self.wtime,
                      ", btime:", self.btime, " stm:", stm,
                      ", time for move (ms):", gs_movetime))

            command = "level 0 " + gs_level
            engine.command(command)
        # fixed search depth
        elif self.type == 4:
            idepth = self.dpth_depth
            if idepth > 39:
                if gv.verbose:
                    print(("search depth (", idepth,
                          ") exceeds max for jcchess engine, setting to 39"))
                idepth = 39
            if gv.verbose:
                print("setting depth for jcchess engine to", idepth)
            engine.depth(idepth)
        # fixed nodes search
        elif self.type == 6:
            inodes = self.nodes_nodes

            if gv.verbose:
                print("setting nodes for jcchess engine to", inodes)
                print("time=", time.time())
            engine.nodes(inodes)
            command = "level 0 59:59"
            engine.command(command)

    def update_gui_time_control(self, stm):
        self.set_toolbar_time_control(self.type, 0, stm)

    def set_toolbar_time_control(self, tc_type, move_time, stm):
        if tc_type == 0:
            self.set_toolbar_time_control0(move_time, stm)
        elif tc_type == 1:
            self.set_toolbar_time_control1(move_time, stm)
        elif tc_type == 2:
            self.set_toolbar_time_control2(move_time, stm)
        elif tc_type == 3:
            self.set_toolbar_time_control3(move_time, stm)
        elif tc_type == 4 or tc_type == 5 or tc_type == 6:
            self.set_toolbar_time_control4(move_time, stm)
        else:
            print("invalid tc type", tc_type)

    # update gui clock for byoyomi TC
    def set_toolbar_time_control0(self, move_time, side_to_move):

        if side_to_move == WHITE:
            time_left = self.wtime
        else:
            time_left = self.btime

        # subtract the move time from the players available time
        new_time = time_left - move_time

        byoyomi = self.byo_byoyomi * 1000

        # If no available time use byoyomi time
        if new_time <= 0:
            hours = 0
            mins = 0
            secs = 0
            if time_left > 0:
                byo = byoyomi - move_time + time_left
            else:
                byo = byoyomi - move_time
            if byo < 0:
                byo = 0
        else:
            secs = int(new_time / 1000)
            hours = int(secs / 3600)
            secs = secs - (hours * 3600)
            mins = int(secs / 60)
            secs = int(secs - (mins * 60))
            byo = 0

        byoyomi = int(byoyomi / 1000)
        byo = int(byo / 1000)

        shours = str(hours)
        if len(shours) < 2:
            shours = "0" + shours

        smins = str(mins)
        if len(smins) < 2:
            smins = "0" + smins

        ssecs = str(secs)
        if len(ssecs) < 2:
            ssecs = "0" + ssecs

        sbyo = str(byo)
        if len(sbyo) < 2:
            sbyo = "0" + sbyo
        sbyoyomi = str(byoyomi)
        if len(sbyoyomi) < 2:
            sbyoyomi = "0" + sbyoyomi

        txt = shours + ":" + smins + ":" + ssecs + "  " + sbyo + "/" + sbyoyomi
        gv.gui.set_toolbar_time_control(txt, side_to_move)

    # update gui clock for classical TC
    def set_toolbar_time_control1(self, move_time, side_to_move):
        if side_to_move == WHITE:
            time_left = self.wtime
            moves_to_go = self.wmoves_to_go
        else:
            time_left = self.btime
            moves_to_go = self.bmoves_to_go

        # subtract the move time from the players available time
        new_time = time_left - move_time

        if new_time <= 0:
            hours = 0
            mins = 0
            secs = 0
        else:
            secs = int(new_time / 1000)
            hours = int(secs / 3600)
            secs = secs - (hours * 3600)
            mins = int(secs / 60)
            secs = int(secs - (mins * 60))

        shours = str(hours)
        if len(shours) < 2:
            shours = "0" + shours

        smins = str(mins)
        if len(smins) < 2:
            smins = "0" + smins

        ssecs = str(secs)
        if len(ssecs) < 2:
            ssecs = "0" + ssecs

        txt = shours + ":" + smins + ":" + ssecs + " " + str(moves_to_go)
        gv.gui.set_toolbar_time_control(txt, side_to_move)

    # update gui clock for incremental TC
    def set_toolbar_time_control2(self, move_time, side_to_move):
        if side_to_move == WHITE:
            time_left = self.wtime
        else:
            time_left = self.btime

        # subtract the move time from the players available time
        new_time = time_left - move_time

        # If no available time use byoyomi time
        if new_time <= 0:
            hours = 0
            mins = 0
            secs = 0
        else:
            secs = int(new_time / 1000)
            hours = int(secs / 3600)
            secs = secs - (hours * 3600)
            mins = int(secs / 60)
            secs = int(secs - (mins * 60))

        shours = str(hours)
        if len(shours) < 2:
            shours = "0" + shours

        smins = str(mins)
        if len(smins) < 2:
            smins = "0" + smins

        ssecs = str(secs)
        if len(ssecs) < 2:
            ssecs = "0" + ssecs

        txt = shours + ":" + smins + ":" + ssecs
        gv.gui.set_toolbar_time_control(txt, side_to_move)

    # update gui clock for fixed time per move TC
    def set_toolbar_time_control3(self, move_time, side_to_move):
        if side_to_move == WHITE:
            time_left = self.wtime
        else:
            time_left = self.btime

        # subtract the move time from the players available time
        new_time = time_left - move_time

        # If no available time use byoyomi time
        if new_time <= 0:
            hours = 0
            mins = 0
            secs = 0
        else:
            secs = int(new_time / 1000)
            hours = int(secs / 3600)
            secs = secs - (hours * 3600)
            mins = int(secs / 60)
            secs = int(secs - (mins * 60))

        shours = str(hours)
        if len(shours) < 2:
            shours = "0" + shours

        smins = str(mins)
        if len(smins) < 2:
            smins = "0" + smins

        ssecs = str(secs)
        if len(ssecs) < 2:
            ssecs = "0" + ssecs

        txt = shours + ":" + smins + ":" + ssecs
        gv.gui.set_toolbar_time_control(txt, side_to_move)

    # update gui clock for fixed search depth TC
    def set_toolbar_time_control4(self, move_time, side_to_move):

        new_time = move_time

        # If no available time use byoyomi time
        if new_time <= 0:
            hours = 0
            mins = 0
            secs = 0
        else:
            secs = int(new_time / 1000)
            hours = int(secs / 3600)
            secs = secs - (hours * 3600)
            mins = int(secs / 60)
            secs = int(secs - (mins * 60))

        shours = str(hours)
        if len(shours) < 2:
            shours = "0" + shours

        smins = str(mins)
        if len(smins) < 2:
            smins = "0" + smins

        ssecs = str(secs)
        if len(ssecs) < 2:
            ssecs = "0" + ssecs

        txt = shours + ":" + smins + ":" + ssecs
        gv.gui.set_toolbar_time_control(txt, side_to_move)



import ez_setup
ez_setup.use_setuptools()
from setuptools import setup

import os
import sys
import shutil

import jcchess.constants

assert sys.version_info >= (3,0)

VERSION = jcchess.constants.VERSION

# linux
data_files=[
  (sys.prefix+'/share/applications',['jcchess.desktop']),
  (sys.prefix+'/share/pixmaps', ['jcchess.png'])]

package_data_list = []

# translations
if shutil.which("msgfmt") is None:
    print("msgfmt not found. Translations will not be built")
else:
    localedir = "locale"
    dirlist = os.listdir(localedir)
    for d in dirlist:
        pth = os.path.join(localedir, d)
        if not os.path.isdir(pth):
            continue
        filein = os.path.join(pth, "LC_MESSAGES", "jcchess.po")
        pthout = os.path.join("jcchess", pth, "LC_MESSAGES")
        if not os.path.exists(pthout):
            try:
                os.makedirs(pthout)
            except OSError as exc:
                print("Unable to create locale folder", pthout)
                sys.exit()
        fileout = os.path.join(pthout, "jcchess.mo")
        os.popen("msgfmt %s -o %s" % (filein, fileout))
        package_data_list.append(os.path.join(pth, "LC_MESSAGES", "jcchess.mo"))

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name="jcchess",
      version=VERSION,
      description="A Chess GUI", 
      include_package_data=True,
      author="John Cheetham",
      author_email="developer@johncheetham.com",
      url="http://www.johncheetham.com/projects/jcchess/",
      long_description=read("README.rst"),
      platforms=["Linux"],
      license="GPLv3+",
      zip_safe=False,

      packages=["jcchess", "chess"],
      package_data={
          "jcchess": package_data_list,
      },
      data_files=data_files,
      entry_points={
          "gui_scripts": [
              "jcchess = jcchess.jcchess:run",
          ]
      },

      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: X11 Applications :: GTK",
          "Intended Audience :: End Users/Desktop",
          "License :: OSI Approved :: GNU General Public License (GPL)",
          "Operating System :: POSIX :: Linux",
          "Programming Language :: Python",
          "Programming Language :: C",
          "Topic :: Games/Entertainment :: Board Games",
          ],

      )



#
#   jcchess 0.0.1
#
#   Copyright (C) 2010-2016 John Cheetham
#
#   web   : http://www.johncheetham.com/projects/jcchess
#   email : developer@johncheetham.com
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GLib
import _thread
import traceback
import os
import platform
import sys
import pickle
import time
from datetime import date

import chess
from . import gv
from . import utils
from . import gui
from . import uci
from . import engine_manager
from . import time_control
from . import set_board_colours
from . import move_list
from . import comments
from . import board
from . import pieces
from . import engine_output
from . import load_save
from .constants import WHITE, BLACK, NEUTRAL, NAME, VERSION


class Game:

    def __init__(self):

        gv.jcchess = self
        self.xname = ""
        Filefound = False
        # set global variables for debug messages
        gv.verbose, gv.verbose_uci, gv.show_moves, gv.show_header = utils.get_verbose()
    
        # prefix to find package files/folders
        self.prefix = utils.get_prefix()
        # set up .jcchess directory in the home directory
        self.jcchesspath = utils.create_settings_dir()
        self.glade_dir = os.path.join(self.prefix, "glade")
        # get settings saved from previous game
        self.settings = utils.get_settings_from_file(self.jcchesspath)
        # Filename from command line
        Startfilename = False
        for arg in range(1, len(sys.argv)):
                 if (sys.argv[arg] != '-v' and sys.argv[arg] != '-vuci' and sys.argv[arg] !=""):
                        if (sys.argv[arg].find(":")!= -1 or sys.argv[arg].startswith("/")) :
                                Startfilename = True
                        if Startfilename == True:
                                self.xname = self.xname + " " + sys.argv[arg] #single file from command-line
                                #print (sys.argv[arg] + "   "+ str(arg) +"    "+ self.xname)
                                if sys.argv[arg].find(".psn") != -1:
                                        Filefound = True
                                        break
                                if sys.argv[arg].find(".gshog") != -1:
                                        Filefound = True
                                        break
        self.xname = self.xname.strip()
        if gv.verbose == True :
                print("Filename read:  " + self.xname + "  File:"+ str(Filefound) +"   "+VERSION)
        self.ask_before_promoting = False
       
        self.gameover = False
        self.time_limit = "00:10"
        self.stopped = True
        self.quitting = False
        self.src = ""
        self.src_x = ""
        self.src_y = ""
        self.startpos = "startpos"
        self.start_stm = WHITE

        self.search_depth = 39
        self.thinking = False
        self.cmove = "none"
        self.movelist = []
        self.lastmove = ""
        self.redolist = []
        self.player = ["Human", "Human"]
        self.pondermove = [None, None]

        opening_book_path = os.path.join(self.prefix, "data/opening.bbk")
        #engine.init(opening_book_path, gv.verbose)

        gv.load_save = load_save.Load_Save()
        gv.gui = gui.Gui()

        gv.pieces = pieces.Pieces()
        gv.pieces.load_pieces(self.prefix)

        # uciw is the instance that plays white
        # ucib is the instance that plays black
        gv.ucib = uci.Uci("b")
        gv.uciw = uci.Uci("w")

        # instantiate board, gui, classes
        gv.tc = time_control.Time_Control()
        gv.engine_manager = engine_manager.Engine_Manager()
        gv.board = board.Board()

        gv.set_board_colours = set_board_colours.Set_Board_Colours(self.prefix)
        # set colours to previous game (if any)
        if self.settings is not None:
            gv.set_board_colours.set_colour_scheme(self.settings.colour_settings)
        gv.gui.build_gui()
        gv.board.build_board()
        self.engine_output = engine_output.get_ref()

        # set level
        #command = "level 0 " + self.time_limit
        #engine.command(command)

        # restore users settings to values from previous game
        self.restore_settings(self.settings)

        gv.ucib.set_engine(self.player[BLACK], None)
        gv.uciw.set_engine(self.player[WHITE], None)
        gv.gui.update_toolbar(self.player)

        self.move_list = move_list.get_ref()

        gv.tc.reset_clock()

        gv.gui.enable_go_button()
        gv.gui.disable_stop_button()

        self.stm = self.get_side_to_move()
        self.timer_active = False
        gv.set_board_colours.apply_colour_settings()
        if Filefound==True:
            gv.lastdir = os.path.dirname(self.xname)
            # single file from command-line
            gv.gui.load_save.load_game_parm(self.xname)

    #
    # Process Human move
    #
    def square_clicked(self, widget, event, data):
        # if in edit board mode then call routine in gui.py to show the edit
        # popup menu
        if gv.gui.get_edit_mode():
            (x, y) = data
            gv.gui.show_edit_popup(event, x, y)
            return

        if self.gameover or self.thinking or self.stopped:
            return

        self.stm = self.get_side_to_move()
        if self.player[self.stm] != "Human":
            return
        gv.gui.set_side_to_move(self.stm)

        # get x,y co-ords of square clicked on (0, 0 is top left)
        (x, y) = data

        # convert the x, y co-ords into the shogi representation
        # (e.g. 8, 6 is 1g)
        sq = gv.board.get_square_posn(x, y)

        # if the square clicked on is a valid source square
        # then set this square as the source square
        if gv.board.valid_source_square(x, y, self.stm):
            self.src = sq
            if gv.verbose:
                print("source square: (x, y) = (", x, ",",  y, ") ", sq)
            self.src_x = x
            self.src_y = y
            self.piece = gv.board.get_piece(x, y)
            # hilite square clicked on
            gv.board.update((chess.square(x, y),))
            return

        # must have a valid source square before checking dest square
        if self.src == "":
            return

        # Not a valid source square, assume destination square
        dst = sq

        move = self.get_move(self.piece, self.src, dst, self.src_x,
                             self.src_y, x, y)
        if move is None:
            return

        # display the move
        self.human_move(move)

    # format human move
    def get_move(self, piece, src, dst, src_x, src_y, dst_x, dst_y):
        if gv.verbose:
            print("in get move")
            print("src=", src)
            print("dst=", dst)
            print("src_x=", src_x)
            print("src_y=", src_y)
            print("dst_x=", dst_x)
            print("dst_y=", dst_y)

        move = self.src + dst
        # check for promotion 
        piece = gv.board.get_piece(src_x, src_y)

        # if moving pawn to end rank then show promotion dialog
        if (piece == 'P' and src[1] == '7' and dst[1] == '8') or \
           (piece == 'p' and src[1] == '2' and dst[1] == '1'):
            testmove = chess.Move.from_uci(move+'q')
            validmove = testmove in gv.board.get_legal_moves()
            if validmove:
                # remove piece during dialog
                #piece = gv.board.remove_piece_at_square(src_x, src_y)
                gv.gui.promote_popup()
                # reinstate piece
                #gv.board.set_piece_at_square(src_x, src_y, piece.piece_type, piece.color)
                move += gv.gui.get_promotion_piece()

        if gv.verbose:
            print("move=", move)
 
        cmove = chess.Move.from_uci(move)
        validmove = cmove in gv.board.get_legal_moves()
        if (not validmove):
            # illegal move
            gv.gui.set_status_bar_msg(_("Illegal Move"))
            return None
        gv.board.add_move(cmove)
        return move

    def human_move(self, move):
        self.movelist.append(move)
        self.redolist = []
        self.lastmove = move

        gv.board.update()

        # update move list in move list window
        self.move_list.update()

        if gv.verbose:
            gv.board.print_board()

        self.src = ""

        self.gameover, msg = self.check_for_gameover()
        if (self.gameover):
            self.stop()
            gv.gui.set_status_bar_msg(msg)
            self.thinking = False
            return

        self.stm = self.get_side_to_move()
        gv.gui.set_side_to_move(self.stm)
        if gv.verbose:
            print("#")
            print("# " + self.get_side_to_move_string(self.stm) + _(" to move"))
            print("#")
        gv.gui.set_status_bar_msg(" ")

        self.src = ""

        # update time for last move
        gv.tc.update_clock()

        gv.gui.apply_drag_and_drop_settings(self.player[self.stm], self.stm)

        if self.player[self.stm] == "Human":

            # set clock ready for move to come
            gv.tc.start_clock(self.stm)

            if not self.timer_active:
                GObject.timeout_add(1000, gv.tc.show_time)

            return

        self.thinking = True
        # disable some functionality while the computer is thinking
        # gv.gui.disable_menu_items()

        # it's the computers turn to move. Kick off a separate thread for
        # computers move so that gui is still useable
        self.ct = _thread.start_new_thread(self.computer_move, ())

        return

    def stop_clicked(self, widget):
        self.stop()

    def stop(self):
        self.stopped = True
        gv.gui.enable_menu_items()
        gv.gui.enable_go_button()
        gv.gui.disable_stop_button()
        gv.gui.unset_all_drag_and_drop_squares()

        # update time
        gv.tc.stop_clock()

        # stop engines
        gv.ucib.stop_engine()
        gv.uciw.stop_engine()
        #engine.movenow()
        gv.gui.set_status_bar_msg(_("stopped"))

    def go_clicked(self, widget):

        # update time control prior to move
        gv.tc.update_gui_time_control(self.stm)

        # side to move
        self.stm = self.get_side_to_move()

        # start a timer to display the time left while the player is thinking
        gv.tc.start_clock(self.stm)

        if not self.timer_active:
            GObject.timeout_add(1000, gv.tc.show_time)

        gv.gui.disable_menu_items()
        gv.gui.disable_go_button()
        gv.gui.enable_stop_button()

        self.stopped = False

        if gv.verbose:
            print("#")
            print("# " + self.get_side_to_move_string(self.stm) + " to move")
            print("#")

        gv.gui.apply_drag_and_drop_settings(self.player[self.stm], self.stm)

        # gv.board.reduce_board_history(self.movelist)
        self.engine_output.clear("w", " ")
        self.engine_output.clear("b", " ")

        if self.player[self.stm] == "Human":
            gv.gui.set_status_bar_msg(_("ready"))
            return

        gv.gui.set_status_bar_msg(_("Thinking ..."))
        # it's the computers turn to move. kick off a separate thread for
        # computers move so that gui is still useable
        self.ct = _thread.start_new_thread(self.computer_move, ())

    def get_prefix(self):
        return self.prefix

    def get_glade_dir(self):
        return self.glade_dir

    def computer_move(self):
        try:
            self.thinking = True

            while True:
                self.thinking = True
                self.stm = self.get_side_to_move()

                # update time for last move
                # gv.tc.update_clock()
                GLib.idle_add(gv.gui.set_side_to_move, self.stm)
                # set clock ready for move to come
                # gv.tc.start_clock(self.stm)

                if gv.verbose:
                    print("#")
                    print(("# " + self.get_side_to_move_string(self.stm) +
                          " to move"))
                    print("#")

                if self.player[self.stm] == "Human":
                    GLib.idle_add(gv.gui.apply_drag_and_drop_settings,
                        self.player[self.stm], self.stm)
                    self.thinking = False
                    gv.tc.start_clock(self.stm)
                    return

                t_start = time.time()

                if self.stm == BLACK:
                    self.uci = gv.ucib
                    # uci_opponent = self.uciw
                    # opponent_stm = WHITE
                else:
                    self.uci = gv.uciw
                    # uci_opponent = self.ucib
                    # opponent_stm = BLACK

                ponder_enabled = gv.engine_manager.get_ponder()

                if not ponder_enabled:
                    # get engines move
                    self.cmove, self.pondermove[self.stm] = self.uci.cmove(
                        self.movelist, self.stm)
                else:
                    # pondering
                    ponderhit = False
                    if (self.pondermove[self.stm] is not None and
                            len(self.movelist) > 0):
                        if (self.movelist[-1] ==
                                self.pondermove[self.stm]):
                            ponderhit = True

                    if ponderhit:
                        # ponderhit, wait for return of bestmove and
                        # pondermove
                        self.cmove, self.pondermove[self.stm] = \
                            self.uci.send_ponderhit(self.stm)
                    else:
                        if self.pondermove[self.stm] is not None:
                            # stop ponder, wait for return of bestmove and
                            # pondermove from ponder
                            bm, pm = self.uci.stop_ponder()
                        # get engines move
                        self.cmove, self.pondermove[self.stm] = \
                            self.uci.cmove(self.movelist, self.stm)

                    # start pondering
                    if self.pondermove[self.stm] is not None:
                        # send position and ponder command,
                        # return immediately
                        self.uci.start_ponder(
                            self.pondermove[self.stm],
                            self.movelist, self.cmove)

                if self.stopped or self.cmove is None:
                    self.thinking = False
                    return

                if gv.verbose:
                    print("computer move is", self.cmove)
                # check if player resigned
                if self.cmove == "resign":
                    if gv.verbose:
                        print("computer resigned")
                    self.gameover = True
                    self.thinking = False
                    colour = self.get_side_to_move_string(self.stm)
                    msg = _("game over - ") + colour + _(" resigned")
                    GLib.idle_add(self.stop)
                    GLib.idle_add(gv.gui.set_status_bar_msg, msg)
                    self.move_list.comments.automatic_comment(self.cmove + _(msg),len(self.movelist))
                    self.thinking = False
                    return

                # engine.setplayer(WHITE)
                #engine.setplayer(self.stm)
                cmove = chess.Move.from_uci(self.cmove)                    
                validmove = cmove in gv.board.get_legal_moves()
                #validmove = engine.hmove(self.cmove)
                if (not validmove):
                    GLib.idle_add(self.stop)
                    GLib.idle_add(gv.gui.set_status_bar_msg,
                        self.cmove + _(" - computer made illegal Move!"))
                    self.gameover = True
                    self.thinking = False
                    self.move_list.comments.automatic_comment(self.cmove + _(" - computer made illegal Move!"),len(self.movelist))
                    return
                gv.board.add_move(cmove)
                if gv.verbose:
                    gv.board.print_board()

                if self.cmove != "":
                    self.movelist.append(self.cmove)
                    self.redolist = []
                    self.lastmove = self.cmove
                else:
                    # empty move is returned by jcchess engine when it is in
                    # checkmate
                    if gv.verbose:
                        print("empty move returned by engine")

                # if program is exitting then quit this thread asap
                if self.quitting:
                    return

                # gv.board.save_board(len(self.movelist))
                # show computer move
                GLib.idle_add(gv.board.update)
                GLib.idle_add(self.move_list.update)

                # if self.player[self.stm] != "jcchess"
                #     and gv.engine_manager.get_ponder():
                #     self.uci.send_ponder()
                #     # self.ctp= thread.start_new_thread(
                #     self.uci.send_ponder, () )

                if gv.verbose:
                    gv.board.print_board()
                    
                if gv.verbose:
                    print("move=", self.cmove)
                msg = self.cmove

                self.gameover, gmsg = self.check_for_gameover()
                if (self.gameover):
                    if (msg == ""):
                        msg = gmsg
                    else:
                        msg = self.get_side_to_move_string(
                            self.stm) + ": " + msg
                        msg = self.cmove + "." + msg + ". " + gmsg
                    self.thinking = False
                    self.stm = self.get_side_to_move()
                    GLib.idle_add(self.stop)
                    GLib.idle_add(gv.gui.set_side_to_move, self.stm)
                    GLib.idle_add(gv.gui.set_status_bar_msg, msg)
                    self.move_list.comments.automatic_comment(self.cmove + _(msg + gmsg),len(self.movelist))
                    return

                #msg = self.get_side_to_move_string(self.stm) + ": " + str(len(self.movelist)) + ". " + msg
                msg = self.get_side_to_move_string(self.stm) + ": " + msg                             
                GLib.idle_add(gv.gui.set_status_bar_msg, msg)
                # keep a sleep in the loop to avoid issues with gui updating
                # in engine v engine games
                time.sleep(0.1)
            self.thinking = False
        except:
            traceback.print_exc()
            return

    def check_for_gameover(self):
        gameover = False
        msg = ""
        #winner = engine.getwinner()
        #if (winner):        
        if gv.board.is_gameover():            
            gameover = True
            #winner -= 1
            #if (winner == BLACK):
            #    msg = _("game over - black wins")
            #elif (winner == WHITE):
            #    msg = _("game over - white wins")
            #elif (winner == NEUTRAL):
            #    msg = _("game over - match drawn")
            #else:
            #    print("invalid value returned from engine getwinner function")
            msg = _("game over")
        return gameover, msg

    def set_promotion_mode(self, mode):
        self.ask_before_promoting = mode

    #
    # Callback Functions
    #

    def quit_game(self, b):
        if gv.verbose:
            gv.testnames()
        self.quit()
        return False

    # This callback quits the program
    def delete_event(self, widget, event, data=None):
        if gv.verbose:
            gv.testnames()
        self.quit()
        return False

    def quit(self):
        self.stopped = True
        self.quitting = True
        #engine.movenow()
        self.save_settings()
        gv.ucib.stop_engine()
        gv.uciw.stop_engine()
        Gtk.main_quit()
        return False

    def quitting(self):
        return self.quitting

    #
    # called from menu by new game and new handicap game
    #
    def new_game_cb(self, action):
        menu_name = action.get_name()
        self.new_game(menu_name)

    # called from new_game_cb in this module and from load_game_psn_from_str
    # in psn.py
    def new_game(self, menu_name):

        self.gameover = False
        #engine.command("new")
        gv.board.init_board()
        if menu_name == "NewGame":
            # Normal Game
            self.startpos = "startpos"
            self.start_stm = WHITE
        else:
            gv.gui.info_box("Error. Invalid menu name:" + menu_name)
            return

        gv.board.update()
        # clean comments (bugfix)
        self.move_list.comments.clear_comments()

        gv.ucib.set_newgame()
        gv.uciw.set_newgame()
        self.movelist = []
        self.redolist = []
        self.lastmove = ""
        # update move list in move list window
        self.move_list.update()
        #clear comment-and move windows
        #self.move_list.comments.automatic_comment("") # just a  Test
        #print("test")
        if gv.show_moves == True:
           start, end =gv.gui.comment_view.get_buffer().get_bounds()
           gv.gui.comment_view.get_buffer().delete(start,end)
           gv.gui.comment_view.get_buffer().insert(start,"-")        
           gv.gui.move_view.get_model().clear()
        gv.gui.set_status_bar_msg("")
        self.stm = self.get_side_to_move()
        gv.gui.set_side_to_move(self.stm)
        gv.tc.reset_clock()
        gv.event = "##"
        gv.gamedate = str(date.today().day) +"."+str(date.today().month)+"."+str(date.today().year)
        gv.gote = self.player[BLACK]
        gv.sente = self.player[WHITE]
        if gv.show_header == True:
         GLib.idle_add(gv.gui.header_lblsente.set_text, "")
         GLib.idle_add(gv.gui.header_lblgote.set_text,"")
         GLib.idle_add(gv.gui.header_lblevent.set_text, "")
         GLib.idle_add(gv.gui.header_lbldate.set_text, "")                
                    
    #
    # save users settings at program termination
    #

    def save_settings(self):

        # get settings
        s = Settings()
        s.name = NAME
        s.version = VERSION
        s.engine_list = gv.engine_manager.get_engine_list()
        s.pieceset = gv.pieces.get_pieceset()
        s.player_white = self.player[WHITE]
        s.player_black = self.player[BLACK]
        s.clock_settings = gv.tc.get_clock_settings()
        s.colour_settings = gv.set_board_colours.get_colour_scheme()
        s.hash_value = gv.engine_manager.get_hash_value()
        s.ponder = gv.engine_manager.get_ponder()
        s.show_coords = gv.gui.get_show_coords()
        s.highlight_moves = gv.gui.get_highlight_moves()
        s.lastdir = gv.lastdir # last dir used
        # Header not saved
        s.event = "##"       #gv.event, mark for not read
        s.gamedate = ""  #gv.gamedate
        s.gote = "" #gv.gote
        s.sente ="" #gv.sente
        if gv.verbose == True:
                print(gv.lastdir + "  saved")
        # pickle and save settings
        try:
            settings_file = os.path.join(self.jcchesspath, "settings")
            f = open(settings_file, "wb")
            pickle.dump(s, f)
            f.close()
        except AttributeError as ae:
            print("attribute error:", ae)
        except pickle.PickleError as pe:
            print("PickleError:", pe)
        except pickle.PicklingError as pe2:
            print("PicklingError:", pe2)
        except Exception as exc:
            print("cannot save settings:", exc)

    #
    # restore users settings at program start-up
    #
    def restore_settings(self, x):
        bepath = None
        if os.name == 'nt':
            # get bundled engine if present
            arch = platform.architecture()[0][0:2]
            bepath = os.path.join(os.path.dirname(self.prefix), "engines", "stockfish_8_x" + arch + ".exe")
            be = ["Stockfish 8 " + arch, bepath, {}]

            # if no settings (first run) add bundled engine and
            # set black player to the engine
            if x is None:
                if os.path.isfile(bepath):
                    gv.engine_manager.set_engine_list([be])
                    self.player[BLACK] = be[0]
                return    
            
        if x:

            # set the engine or human for each player
            try:
                self.player[WHITE] = x.player_white
                self.player[BLACK] = x.player_black
            except Exception as e:
                if gv.verbose:
                    print(e, ". player setting not restored")

            # engine list
            try:
                if x.engine_list:
                    # if old format of [(engine, path), ...] 
                    # then reformat to [[engine, path, {}], ...]                    
                    if len(x.engine_list[0]) != 3:
                        print("reformatting engine list")
                        new_engine_list = []
                        for olde in x.engine_list:
                            newe = [olde[0], olde[1], {}]
                            new_engine_list.append(newe)
                        x.engine_list = new_engine_list
                # if bundled engine add it to the list
                # if it is not already there
                if bepath is not None and os.path.isfile(bepath):
                    found = False
                    for e in x.engine_list:
                        if e[0] == be[0]:
                            found = True
                            break
                    if not found:
                        # if no engines there already set black
                        # player to newly added bundled engine
                        if not x.engine_list:
                            self.player[BLACK] = be[0]
                        x.engine_list.insert(0, be)
                gv.engine_manager.set_engine_list(x.engine_list)
            except Exception as e:
                if gv.verbose:
                    print(e, ". engine list not restored")

            try:
                gv.gamedate = x.gamedate
                if gv.verbose == True:
                    print("date read: " + gv.gamedate)
            except Exception as e:
                if gv.verbose:
                    print (e, ". date not restored")

            # pieceset
            try:
                gv.pieces.set_pieceset(x.pieceset)
            except Exception as e:
                if gv.verbose:
                    print(e, ". pieceset setting not restored")

            # time controls
            try:
                cs = x.clock_settings
                gv.tc.restore_clock_settings(cs)
            except Exception as e:
                if gv.verbose:
                    print(e, ". time controls not restored")

            # hash value
            try:
                hash_value = x.hash_value
                gv.engine_manager.set_hash_value(hash_value)
            except Exception as e:
                if gv.verbose:
                    print(e, ". hash value not restored")

            # ponder (true/false)
            try:
                ponder = x.ponder
                gv.engine_manager.set_ponder(ponder)
            except Exception as e:
                if gv.verbose:
                    print(e, ". ponder not restored")

            # show coordinates (true/false)
            try:
                show_coords = x.show_coords
                gv.gui.set_show_coords(show_coords)
            except Exception as e:
                if gv.verbose:
                    print(e, ". show_coords not restored")

            # highlight moves (true/false)
            try:
                highlight_moves = x.highlight_moves
                gv.gui.set_highlight_moves(highlight_moves)
            except Exception as e:
                if gv.verbose:
                    print(e, ". highlight_moves not restored")
            # lastdir
            try: 
                gv.lastdir = x.lastdir
                if gv.verbose == True:
                        print("path read: " + gv.lastdir)
            except Exception as e:
                if gv.verbose: 
                        print (e, ". lastdir not restored")
            #header : werden nicht gelesen
            gv.gamedate = ""
            gv.event = "##"
            gv.gote = ""
            gv.sente = ""

    # called from this module and from comments.py
    def convert_move(self, moveno):
        if moveno % 2 == 1:
            strmoveno = str(int((moveno + 1) / 2)) + "."
        else:
            strmoveno = str(int((moveno + 1) / 2)) + "..." 
        return strmoveno
        
    def goto_move(self, move_idx):
        try:
            gv.ucib.stop_engine()
            gv.uciw.stop_engine()
        except:
            pass
        self.gameover = False
        if move_idx < len(self.movelist):
            while move_idx < len(self.movelist):
                self.undo_move()
        else:
            while move_idx > len(self.movelist):
                self.redo_move()

        self.stm = self.get_side_to_move()
        gv.gui.set_side_to_move(self.stm)

        gv.board.update()
        # gv.gui.set_status_bar_msg(" ")
        move = None
        try:
            move = self.movelist[len(self.movelist) - 1]
            self.lastmove = move
            # print "move ",move
        except IndexError:
            pass

        if move is not None:
            gv.gui.set_status_bar_msg(self.convert_move(move_idx)+move)
            if gv.show_moves == True:
               nmoves = len(self.movelist)
               path =(nmoves-1,)
               sel = gv.gui.move_view.get_selection()
               sel.select_path(path)  # path has to be set to nmoves-1 to hit the right selection
               self.move_list.set_move(nmoves)
               #print(nmoves, "goto move")             
        else:
            gv.gui.set_status_bar_msg(" ")

    #
    # called from gui.py when undo button click on toolbar (passed widget is
    # Gtk.ToolButton object) and when undo move is selected from menu
    # (or ctrl-u is pressed) (passed widget is Gtk.Action object)
    #
    def undo_single_move(self, b):
        
        move = None
        m=None
        nmove = len(self.movelist)
        try:
            move = self.movelist.pop()
            m=gv.board.remove_move()
            self.redolist.append(move)
            self.lastmove = move
            self.stm = self.get_side_to_move()
            gv.gui.set_side_to_move(self.stm)
        except IndexError:
            pass

        try:
            gv.ucib.stop_engine()
            gv.uciw.stop_engine()
        except:
            pass
        self.gameover = False
        #gv.board.update((m.from_square, m.to_square))
        # set move list window to last move
        self.move_list.set_move(len(self.movelist))
        self.goto_move(len(self.movelist))
        if m is not None:
            squares_to_hilite=(m.from_square, m.to_square)
        else:
            squares_to_hilite=None
        gv.board.update(squares_to_hilite)
        if move is not None:
            gv.gui.set_status_bar_msg("back: (" + self.convert_move(nmove) + move + ")")
            #gv.gui.set_status_bar_msg("back: (" + move + ")")

    # undo a move without updating the gui
    def undo_move(self):
        gv.board.remove_move()
        move = None
        try:
            move = self.movelist.pop()
            self.redolist.append(move)
                  
        except IndexError:
            pass

    def undo_all(self, toolbutton):
        try:
            gv.ucib.stop_engine()
            gv.uciw.stop_engine()
        except:
            pass
        self.gameover = False
        while len(self.movelist) != 0:
            self.undo_move()
        self.lastmove = ""

        self.stm = self.get_side_to_move()
        gv.gui.set_side_to_move(self.stm)

        gv.board.update()
        # set move list window to initial position
        self.move_list.set_move(1)
        gv.gui.set_status_bar_msg(" ")
        if gv.show_moves == True:
            start, end =gv.gui.comment_view.get_buffer().get_bounds()
            gv.gui.comment_view.get_buffer().delete(start,end)
            gv.gui.comment_view.get_buffer().insert(start,"-")        
                  


    #
    # called from gui.py when redo button click on toolbar (passed widget is
    # Gtk.ToolButton object) and when redo move is selected from menu
    # (or ctrl-r is pressed) (passed widget is Gtk.Action object)
    #
    def redo_single_move(self, widget):
        move = None
        try:
            move = self.redolist.pop()
            # get side to move before appending to movelist
            self.stm = self.get_side_to_move()
            self.movelist.append(move)
            self.lastmove = move

            # do the move in jcchess engine
            #engine.setplayer(self.stm)
            #engine.hmove(move)
            cmove = chess.Move.from_uci(move)
            gv.board.add_move(cmove)

            # side to move changes to opponent
            self.stm = self.get_side_to_move()
            gv.gui.set_side_to_move(self.stm)
                    
        except IndexError:
            pass

        try:
            gv.ucib.stop_engine()
            gv.uciw.stop_engine()
        except:
            pass
        gv.board.update()
        # set move list window to last move
        self.move_list.set_move(len(self.movelist))
        self.goto_move(len(self.movelist))
        nmove = len(self.movelist)
        if move is not None:
           gv.gui.set_status_bar_msg("forward: (" + self.convert_move(nmove) + move + ")")
           #gv.gui.set_status_bar_msg("forward: (" + move + ")")
            
    # redo a move without updating the gui
    def redo_move(self):
        move = None
        try:
            move = self.redolist.pop()

            # get side to move before appending to movelist
            self.stm = self.get_side_to_move()
            self.movelist.append(move)
            
            # do the move in jcchess engine
            #engine.setplayer(self.stm)
            #engine.hmove(move)
            cmove = chess.Move.from_uci(move)
            gv.board.add_move(cmove)
        except IndexError:
            pass

    def redo_all(self, toolbutton):
        while len(self.redolist) != 0:
            self.redo_move()
        self.stm = self.get_side_to_move()
        gv.gui.set_side_to_move(self.stm)
        gv.board.update()
        # set move list window to last move
        self.move_list.set_move(len(self.movelist))                         
        self.goto_move(len(self.movelist))
        if gv.show_moves == True:
         startc, endc =gv.gui.comment_view.get_buffer().get_bounds()
         gv.gui.comment_view.get_buffer().delete(startc,endc)
         gv.gui.comment_view.get_buffer().insert(startc,"-")   
        if len(self.movelist) != 0 and gv.show_moves:
           gv.gui.move_view.get_selection().select_path(str(len(self.movelist)-1))
        nmove = len(self.movelist)
        move = None
        try:
            move = self.movelist[len(self.movelist) - 1]
        except IndexError:
            pass            

        if move is not None:
            gv.gui.set_status_bar_msg(" (" + self.convert_move(nmove) + move + ")")
            #gv.gui.set_status_bar_msg(" (" + move + ")")
            self.lastmove = move

    def set_movelist(self, movelist):
        self.movelist = movelist

    def set_redolist(self, redolist):
        self.redolist = redolist

    def set_startpos(self, startpos):
        self.startpos = startpos

    def set_side_to_move(self, stm):
        self.stm = stm

    def get_stm(self):
        return self.stm

    def get_move_count(self):
        return len(self.movelist) + 1

    def get_side_to_move(self):

        # get side to move for the first move of the game
        # This is white for standard start position but may
        # be different with a fen
        start_stm = self.get_stm_from_fen(self.startpos)

        if len(self.movelist) % 2 == 0:
            return start_stm
        else:
            return start_stm ^ 1

    def get_side_to_move_string(self, stm):
        if stm == BLACK:
            return _("black")
        else:
            return _("white")

    def get_stm_from_fen(self, fen):
        if fen == "startpos":
            # normal game
            stm = WHITE
        else:
            sp = fen.split()
            try:
                if sp[1] == "w":
                    stm = WHITE
                else:
                    stm = BLACK
            except:
                    stm = BLACK
        return stm

    def get_player(self, side):
        return self.player[side]
        
    def set_player(self, side, player):
        self.player[side] = player
        gv.ucib.set_engine(self.player[side], None)
        gv.gui.update_toolbar(self.player)

    def move_now(self, b):
        if self.player[self.stm] == "Human":
            return

        if not self.thinking:
            return

        # UCI engine
        try:
            if self.stm == BLACK:
                gv.ucib.command("stop\n")
            else:
                gv.uciw.command("stop\n")
        except:
            pass

    def get_movelist(self):
        return self.movelist

    def get_redolist(self):
        return self.redolist

    def set_lastmove(self, lastmove):
        self.lastmove = lastmove

    def set_players(self, b):
        dialog = Gtk.Dialog(
            _("Players"), gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OK, Gtk.ResponseType.OK))
        dialog.set_title(_("Set Players"))

        elist = gv.engine_manager.get_engine_list()

        # White Player
        fr = Gtk.Frame.new(_("White"))
        dialog.vbox.pack_start(fr, True, True, 15)
        # dialog.vbox.set_spacing(20)
        vb = Gtk.VBox(False, 0)
        fr.add(vb)

        comboboxw = Gtk.ComboBoxText()
        comboboxw.append_text(_("Human"))

        if self.player[WHITE] == "Human":
            comboboxw.set_active(0)

        i = 1
        for (engine_name, path, ucioptions) in elist:
            comboboxw.append_text(engine_name)

            if engine_name == self.player[WHITE]:
                comboboxw.set_active(i)
            i += 1
        vb.pack_start(comboboxw, True, True, 15)

        # Black Player
        fr = Gtk.Frame.new(_("Black"))
        dialog.vbox.pack_start(fr, True, True, 15)
        vb = Gtk.VBox(False, 0)
        fr.add(vb)

        comboboxb = Gtk.ComboBoxText()
        comboboxb.append_text(_("Human"))
        if self.player[BLACK] == "Human":
            comboboxb.set_active(0)

        i = 1
        for (engine_name, path, ucioptions) in elist:
            comboboxb.append_text(engine_name)
            if engine_name == self.player[BLACK]:
                comboboxb.set_active(i)
            i += 1
        vb.pack_start(comboboxb, True, True, 15)

        dialog.show_all()

        # If user hasn't clicked on OK then exit now
        if dialog.run() != Gtk.ResponseType.OK:
            dialog.destroy()
            return

        self.player[BLACK] = comboboxb.get_active_text()
        self.player[WHITE] = comboboxw.get_active_text()

        gv.ucib.set_engine(self.player[BLACK], None)
        gv.uciw.set_engine(self.player[WHITE], None)

        gv.gui.update_toolbar(self.player)

        dialog.destroy()

    def get_startpos(self):
        return self.startpos

    def get_stopped(self):
        return self.stopped


# class to save settings on program exit and restore on program start
class Settings:
    pass


def run():
    Game()
    GObject.threads_init()
    Gtk.main()
    return 0

if __name__ == "__main__":
    run()



#
#   engine_manager.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
import os

from . import uci
from . import utils
from .constants import WHITE, BLACK
from . import gv


class Engine_Manager:

    def __init__(self):
        # engine list is list of (name, path) pairs
        self.engine_list = []
        self.glade_file = None
        self.hash_value = 256
        self.ponder = False
        # bundled engines directory
        self.bedir = os.path.join(os.path.dirname(utils.get_prefix()), "engines")

    def common_settings(self, b):

        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir,
                                       "common_engine_settings.glade")
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        dialog = self.builder.get_object("common_engine_settings")
        dialog.set_transient_for(gv.gui.get_window())
        # ponder check button
        checkbutton = self.builder.get_object("ponderbutton")
        checkbutton.set_active(self.ponder)

        # hash value
        adj = self.builder.get_object("adjustment1")
        adj.set_value(self.hash_value)
        adj.set_lower(0.00)
        adj.set_upper(10000.00)
        adj.set_step_increment(1)
        adj.set_page_increment(10)
        adj.set_page_size(0)

        response = dialog.run()

        resp_cancel = 1
        resp_ok = 2
        if response == resp_ok:
            self.hash_value = int(adj.get_value())
            self.ponder = checkbutton.get_active()

        dialog.destroy()

    def get_ponder(self):
        return self.ponder

    def set_ponder(self, ponder):
        self.ponder = ponder

    def get_hash_value(self):
        return self.hash_value

    def set_hash_value(self, hash_value):
        self.hash_value = hash_value

    def engines(self, b):

        engine_list = self.get_engine_list()

        dialog = Gtk.Dialog(
            _("Engines"), gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK,
             Gtk.ResponseType.OK)
        )

        hb = Gtk.HBox(False, 20)
        hb.show()
        
        # left side buttons
        lbb = Gtk.VButtonBox()
        lbb.set_layout(Gtk.ButtonBoxStyle.START)
        lbb.show()

        al = Gtk.Alignment.new(xalign=0.0, yalign=0.5, xscale=0.0, yscale=0.0)
        al.add(lbb)
        al.show()

        buttons = (
            _("Move to Top"),
            _("Move Up"),
            _("Move Down"),
            _("Move to Bottom")
            )
        for label in buttons:            
            button = Gtk.Button(label)
            button.show()
            lbb.add(button)
            button.connect("clicked", self.move_engine, "move engine")

        hb.pack_start(al, False, True, 10)
        
        # scroll window for list of engines
        sw = Gtk.ScrolledWindow.new(None, None)
        sw.show()
        fr = Gtk.Frame()
        fr.show()
        liststore = Gtk.ListStore(str, str)
        for e in engine_list:
            engine_name, path, ucioptions = e
            #liststore.append(e)
            liststore.append((engine_name, path))
        self.liststore = liststore

        treeview = Gtk.TreeView(liststore)
        self.treeview = treeview
        tvcolumn = Gtk.TreeViewColumn(_("Select an Engine:"))
        treeview.append_column(tvcolumn)
        cell = Gtk.CellRendererText()
        # cell.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
        cell.set_property("cell-background-gdk", Gdk.color_parse("#F8F8FF"))
        tvcolumn.pack_start(cell, True)
        tvcolumn.set_min_width(200)
        tvcolumn.set_attributes(cell, text=0)
        treeview.show()
        sw.add(treeview)
        fr.add(sw)
        hb.pack_start(fr, True, True, 10)

        # right side buttons
        treeview.connect("button-press-event", self.engine_changed)
        bb = Gtk.VButtonBox()
        bb.set_layout(Gtk.ButtonBoxStyle.START)
        bb.show()

        al = Gtk.Alignment.new(xalign=0.0, yalign=0.5, xscale=0.0, yscale=0.0)
        al.add(bb)
        al.show()

        add_button = Gtk.Button(_("Add"))
        add_button.show()
        bb.add(add_button)
        add_button.connect("clicked", self.add_engine, "add engine")

        self.delete_button = Gtk.Button(_("Delete"))
        self.delete_button.set_sensitive(False)
        self.delete_button.show()
        bb.add(self.delete_button)
        self.delete_button.connect("clicked",
                                   self.delete_engine, "delete engine")

        self.rename_button = Gtk.Button(_("Rename"))
        self.rename_button.set_sensitive(False)
        self.rename_button.show()
        bb.add(self.rename_button)
        self.rename_button.connect("clicked",
                                   self.rename_engine, "rename engine")
                                   
        self.configure_button = Gtk.Button(_("Configure"))
        self.configure_button.show()
        bb.add(self.configure_button)
        self.configure_button.connect("clicked",
                             self.configure_engine, "configure engine")

        hb.pack_start(al, False, True, 10)

        dialog.vbox.pack_start(hb, True, True, 15)

        # set size of scrollwindow
        minsize, naturalsize = treeview.get_preferred_size()
        sw.set_size_request(minsize.width+20, 300)
        
        dialog.set_default_response(Gtk.ResponseType.OK)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            # rebuild engine list
            elist = []
            tm = self.treeview.get_model()
            l_iter = tm.get_iter_first()
            while (l_iter is not None):
                name = tm.get_value(l_iter, 0)
                path = tm.get_value(l_iter, 1)                
                elist.append([name, path, self.get_uservalues(name)])
                l_iter = tm.iter_next(l_iter)
            self.set_engine_list(elist)
        dialog.destroy()

    def engine_changed(self, widget, event):
        GObject.idle_add(self.engine_changed2)

    def engine_changed2(self):
        name, path = self.get_selected_engine()
        # don't allow rename/delete of bundled engines
        if os.path.normpath(self.bedir) in os.path.normpath(path):
            self.delete_button.set_sensitive(False)
            self.rename_button.set_sensitive(False)
        else:
            self.delete_button.set_sensitive(True)
            self.rename_button.set_sensitive(True)

    def get_selected_engine(self):
        ts = self.treeview.get_selection()
        # get liststore object/iter
        lso, l_iter = ts.get_selected()
        tm = self.treeview.get_model()
        try:
            name = tm.get_value(l_iter, 0)
            path = tm.get_value(l_iter, 1)
        except:
            l_iter = tm.get_iter_first()
            ts.select_iter(l_iter)
            name = tm.get_value(l_iter, 0)
            path = tm.get_value(l_iter, 1)
        return name, path

    # move engine up the list by 1 position
    def move_engine(self, widget, data=None):

        # get iter of selected engine
        ts = self.treeview.get_selection()
        # get liststore object/iter
        lso, l_iter = ts.get_selected()
        tm = self.treeview.get_model()
        if l_iter is None:
            gv.gui.info_box(_("no engine selected"))
            return

        if widget.get_label() == _("Move to Top"):
            self.liststore.move_after(l_iter, None)
        elif widget.get_label() == _("Move Up"):
            iter_prev = tm.iter_previous(l_iter)
            if iter_prev is not None:
                self.liststore.swap(l_iter, iter_prev)
        elif widget.get_label() == _("Move Down"):
            iter_next = tm.iter_next(l_iter)
            if iter_next is not None:
                self.liststore.swap(l_iter, iter_next)
        elif widget.get_label() == _("Move to Bottom"):
            self.liststore.move_before(l_iter, None)

    def add_engine(self, widget, data=None):
        dialog = Gtk.FileChooserDialog(
            _("Add UCI Engine.."),
            gv.gui.get_window(),
            Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        )
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_current_folder(os.path.expanduser("~"))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            fname = dialog.get_filename()
            # print "attempting to add new engine"
            path = dialog.get_filename()
            # use a new uci object so not to mess with main game engine
            u = uci.Uci("0")
            u.set_engine("addengine", path)
            errmsg, name = u.test_engine(path)
            if errmsg != "":
                err_dialog = Gtk.Dialog(_("Error Adding Engine"), None, 0,
                                        (Gtk.STOCK_OK, Gtk.ResponseType.OK))
                err_dialog.set_default_size(380, -1)
                lbl = Gtk.Label(label=errmsg)
                lbl.show()
                al = Gtk.Alignment.new(xalign=0.0, yalign=0.5,
                                       xscale=0.0, yscale=0.0)
                al.set_padding(20, 20, 20, 20)
                al.add(lbl)
                al.show()
                err_dialog.vbox.pack_start(al, True, False, 0)
                response = err_dialog.run()
                err_dialog.destroy()
            else:
                self.liststore.append([name, path])
                self.add_engine_to_list([name, path, {}])

        dialog.destroy()

    # Called from this module by clicking 'configure' button
    # or from gui.py by selecting configure engine from menu
    def configure_engine(self, widget, data=None):
        # button was clicked
        if widget.get_name() == "GtkButton":
            ts = self.treeview.get_selection()

            # get liststore object/iter
            lso, l_iter = ts.get_selected()
            tm = self.treeview.get_model()

            if l_iter is None:
                gv.gui.info_box(_("no engine selected"))
                return

            name = tm.get_value(l_iter, 0)
            path = tm.get_value(l_iter, 1)
            u = uci.Uci("1")
            u.set_engine(name, path)
            #u.start_engine(path)
            u.UCI_options()
            return

        # configure engine menu item was selected    
        if widget.get_name() == "ConfigureEngine1":
            player = WHITE
            uciref = gv.uciw
        else:
            player = BLACK
            uciref = gv.ucib

        # If not an engine return
        if gv.jcchess.get_player(player) == "Human":
            gv.gui.info_box(_("No options to configure"))
            return
            
        uciref.UCI_options()

    def delete_engine(self, widget, data=None):

        ts = self.treeview.get_selection()

        # get liststore object/iter
        lso, l_iter = ts.get_selected()
        tm = self.treeview.get_model()

        if l_iter is None:
            gv.gui.info_box(_("no engine selected"))
            return

        name = tm.get_value(l_iter, 0)
        path = tm.get_value(l_iter, 1)

        if not lso.remove(l_iter):
            # set to 1st engine after a delete if iter no longer valid
            l_iter = tm.get_iter_first()

        if l_iter is not None:
            ts.select_iter(l_iter)
            
        # if the deleted engine is in use as one of the players
        # then change that player to human    
        for side in (WHITE, BLACK):
            if name == gv.jcchess.get_player(side):
                gv.jcchess.set_player(side, "Human")

    def rename_engine(self, widget, data=None):

        ts = self.treeview.get_selection()

        # get liststore object/iter
        lso, l_iter = ts.get_selected()
        tm = self.treeview.get_model()

        if l_iter is None:
            gv.gui.info_box(_("no engine selected"))
            return

        name = tm.get_value(l_iter, 0)
        path = tm.get_value(l_iter, 1)

        dialog = Gtk.MessageDialog(
            None,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.OK_CANCEL,
            None)

        markup = "<b>" + _("Rename Engine") + "</b>"
        dialog.set_markup(markup)

        # create the text input fields
        entry = Gtk.Entry()
        entry.set_text(name)
        entry.set_max_length(30)
        entry.set_width_chars(30)

        tbl = Gtk.Table(1, 2, True)
        tbl.attach(Gtk.Label(label=_("Engine Name: ")), 0, 1, 0, 1)
        tbl.attach(entry, 1, 2, 0, 1)

        dialog.vbox.add(tbl)

        dialog.show_all()

        # If user hasn't clicked on OK then exit now
        if dialog.run() != Gtk.ResponseType.OK:
            dialog.destroy()
            return

        # user clicked OK so update with the values entered
        newname = entry.get_text()
        lso.set_value(l_iter, 0, newname)
        dialog.destroy()

    def get_path(self, engine_name):
        for (ename, path, ucioptions) in self.engine_list:
            if ename == engine_name:
                return path
        return None

    def get_uservalues(self, engine_name):
        for (ename, path, ucioptions) in self.engine_list:
            if ename == engine_name:
                return ucioptions
        return {}

    def set_uservalues(self, engine_name, options):
        i = 0
        for (ename, path, ucioptions) in self.engine_list:
            if ename == engine_name:
                newdict = {}
                for option in options:
                    name = option[0]
                    default = option[2]
                    value = option[6]
                    if value != default:
                        newdict[name] =  value
                self.engine_list[i][2] = newdict
            i += 1

    def add_engine_to_list(self, engine_data):
        self.engine_list.append(engine_data)

    def get_engine_list(self):
        return self.engine_list

    def set_engine_list(self, engine_list):
        self.engine_list = engine_list



#
#   gamelist.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
import os
from . import gv


class Gamelist:

    gamelist_ref = None

    def __init__(self):

        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "gamelist.glade")

        # create gamelist window
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("gamelist_window")
        self.treeview = self.builder.get_object("gamelist_treeview")
        self.liststore = self.builder.get_object("liststore1")

        cell0 = Gtk.CellRendererText()
        # cell0.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
        tvcolumn0 = Gtk.TreeViewColumn()
        self.treeview.append_column(tvcolumn0)
        tvcolumn0.pack_start(cell0, True)
        tvcolumn0.set_min_width(50)
        tvcolumn0.set_attributes(cell0, text=0)
        self.treeview.connect("row_activated", self.entry_clicked)
        self.tree_selection = self.treeview.get_selection()

        self.window.hide()

    # user has closed the window
    # just hide it
    def delete_event(self, widget, event):
        self.window.hide()
        return True  # do not propagate to other handlers

    def entry_clicked(self, widget, data=None, data2=None):
        button = None
        self.loadgame_button_clicked_cb(button)
        
    # called from gui.py when doing view gamelist
    def show_gamelist_window_cb(self, action):
        self.show_gamelist_window()

    # called from psn.py when opening multi-game file
    # and from show_gamelist_window_cb above
    def show_gamelist_window(self):
        # "present" will show the window if it is hidden
        # if not hidden it will raise it to the top
        self.window.present()
        return

    def clear(self):
        self.liststore.clear()
        
    def addgame(self, gameno, headers):
        h = str(gameno) + ". "
        hdrno = 1
        for header in headers:
            header_value = headers.get(header)
            # ignore unset headers
            if "?" in header_value:
                continue
            if hdrno > 1:
                h += ", "
            h += header + ": " + header_value
            hdrno += 1
        self.liststore.append([h])

    def loadgame_button_clicked_cb(self, button):
        (treemodel, treeiter) = self.tree_selection.get_selected()
        if treeiter is not None:
            game_str = treemodel.get_value(treeiter, 0)
            gameno = ""
            i = 0
            while game_str[i] != ".":
                gameno += game_str[i]
                i += 1
            try:
                gameno = int(gameno)
            except ValueError as ve:
                return
            gv.load_save.load_game_from_multigame_file(gameno)


def get_ref():
    if Gamelist.gamelist_ref is None:
        Gamelist.gamelist_ref = Gamelist()
    return Gamelist.gamelist_ref



#
#   engine_debug.py - Display USI Engine Debug Window
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import GObject
import os

from .constants import WHITE, BLACK
from . import gv


class Engine_Debug:

    engine_debug_ref = None

    def __init__(self):
        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "engine_debug.glade")
        Engine_Debug.engine_debug_ref = self

        self.debug_text = ""
        self.window = None

    def clear_text(self, b):
        self.tb.set_text("")

    # send command to engine1 (white)
    def engine1_button_clicked(self, b):

        player = gv.jcchess.get_player(WHITE)
        if player == "Human" or player == "jcchess":
            GObject.idle_add(
                self.add_to_log,
                "# command not sent - player 1 (white) is not a USI engine")
            return

        cmd = self.cmd_entry.get_text() + "\n"
        gv.usiw.command(cmd)

    # send command to engine2 (black)
    def engine2_button_clicked(self, b):

        player = gv.jcchess.get_player(BLACK)
        if player == "Human" or player == "jcchess":
            GObject.idle_add(
                self.add_to_log,
                "# command not sent - player 2 (black) is not a USI engine")
            return

        cmd = self.cmd_entry.get_text() + "\n"
        gv.usib.command(cmd)

    # user has closed the window
    # just hide it
    def delete_event(self, widget, event):
        self.window.hide()
        return True  # do not propagate to other handlers

    def add_to_log(self, msg):
        msg = msg + "\n"
        try:
            # append to end of buffer
            end_iter = self.tb.get_end_iter()
            self.tb.insert(end_iter, msg)
            # scroll to end
            GObject.idle_add(self.scroll_to_end)
        except AttributeError:
            # engine debug window has not been opened. Append the debug
            # messages until it is opened
            self.debug_text += msg

    def show_debug_window(self, b):

        # window already exists and is hidden so just show it
        if self.window is not None:
            # "present" will show the window if it is hidden
            # if not hidden it will raise it to the top
            self.window.present()
            return

        # This is the first time the user has opened the engine debug
        # window so need to create it.
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("engine_debug_window")
        self.tv = self.builder.get_object("engine_debug_textview")
        self.tv.set_editable(False)
        self.tb = self.tv.get_buffer()
        self.tb.set_text(self.debug_text)
        self.debug_text = ""

        # used to type commands and send them to the engine
        self.cmd_entry = self.builder.get_object("engine_debug_entry")

        self.window.show_all()

        # scroll to end
        GObject.idle_add(self.scroll_to_end)

    def scroll_to_end(self):
        within_margin = 0.2
        end_iter = self.tb.get_end_iter()
        self.tv.scroll_to_iter(end_iter, within_margin, False, 0.5, 0.5)


def get_ref():
    if Engine_Debug.engine_debug_ref is None:
        Engine_Debug.engine_debug_ref = Engine_Debug()
    return Engine_Debug.engine_debug_ref
